VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "Winsock"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
'===========================================================================
'
' Winsock Class
'
' This class is a drop-in replacement to the Winsock control (MSWINSCK.OCX).
' Unlike the Winsock control, this class does NOT require a hosted form, and
' thus does not require a control array to Accept incoming connection
' requests. There are additional enhancements over the Winsock control
' interface, such as early-bound events (including events for arrays and
' collections of Winsock instances); optional quasi-synchronous non-blocking
' calls, and various other options providing more control over the socket
' communication process.
'
' Since this class mimics the Winsock control's functional interface almost
' exactly, you are able to use it in place of the Winsock control with
' little or no code changes.
'
' 04/10/2002 1. Broke interface to supply a default of -1 in the DataLen
'               parm of the SendData method, and to change the TimeoutValue
'               property to the SendTimeout and ReceiveTimeout properties.
'               Accordingly, the library was renamed from Winsock to
'               kvbWinsock.
'            2. Fixed bug w/Synchronous param in SendData method.
'            3. Fixed bug w/LocalIP property.
' 04/25/2002 1. Fixed failure to send/recv for Listening sockets.
'            2. Fixed crash when recv data into byte array.
' 04/27/2002 1. Added the KillerVB Live! Chat sample application.
'===========================================================================
'
' Author:             Monte Hansen [monte@killervb.com]
' Dependencies:       M+Winsock.Bas, M+Common.Bas & I+WinsockEvents.Cls
'
'===========================================================================
'
' * Copyright  1999-2002 by Monte Hansen, All Rights Reserved Worldwide *
'
' Monte Hansen (The  Author)  grants a royalty-free right to use,  modify,
' and distribute this code (The Code) in compiled form,  provided that you
' agree that The Author has no warranty,  obligations or liability for The
' Code.  You may distribute  The Code  among peers but may not sell it, or
' distribute  it  on  any  electronic  or  physical media such  as  floppy
' diskettes,  compact disks,  bulletin boards,  web  sites,  and the like,
' without first obtaining The Author's consent.
'
' When distributing  The Code  among peers,  it is  respectfully requested
' that it be distributed  as is,  but at  no time  shall it be distributed
' without the copyright notice hereinabove.
'
' This license is subject to change without notice. The current license can
' be found at this address: http://www.killervb.com/kvbLicense.aspx
'
' A credit would be nice if you use The Code ;>
'
'===========================================================================
Option Explicit

'===========================================================================
'   Constants
'===========================================================================
Private Const Module                As String = "Winsock"

Private Const INFINITE              As Long = &HFFFFFFFF  ' Infinite timeout
Private Const INVALID_HANDLE_VALUE  As Long = -1
Private Const ERROR_IO_PENDING      As Long = 997
Private Const HEAP_NO_SERIALIZE     As Long = &H1&
Private Const HEAP_ZERO_MEMORY      As Long = &H8&
Private Const WS_POPUP              As Long = &H80000000
Private Const WM_CLOSE              As Long = &H10&

'===========================================================================
'   Enums
'===========================================================================
Public Enum SocketProtocols
  sckTCPProtocol = 0
  sckUDPProtocol = 1
' Not yet supported:
' sckIrDA = 2
End Enum

' Strings can be sent in Unicode or ANSI:
Public Enum TextModes
  tmAnsi = 0
  tmUnicode = 1
End Enum

Public Enum SocketStates
  sckClosed = 0
  sckOpen = 1
  sckListening = 2
  sckConnectionPending = 3
  sckResolvingHost = 4
  sckHostResolved = 5
  sckConnecting = 6
  sckConnected = 7
  sckClosing = 8
  ' This for compatability:
  sckError = 9
End Enum


Public Enum WinsockErrors
  ' Most of these are defined only to be compatible with
  ' the MS Winsock control.

  'Constant/Value                 Description
  ' -- VB errors raised
  sckOutOfMemory = 7              ' Out of memory
  sckInvalidPropertyValue = 380   ' The property value is invalid.
  sckGetNotSupported = 394        ' The property can't be read.
  sckSetNotSupported = 383        ' The property is read-only.
  sckOpCanceled = 1004            ' The operation was canceled.
  ' -- Winsock control related errors
  sckBadState = 40006             ' Wrong protocol or connection state for the requested transaction or request.
  sckInvalidArg = 40014           ' The argument passed to a function was not in the correct format or in the specified range.
  sckSuccess = 40017              ' Successful.
  sckUnsupported = 40018          ' Unsupported variant type.
  sckInvalidOp = 40020            ' Invalid operation at current state
  sckOutOfRange = 40021           ' Argument is out of range.
  sckWrongProtocol = 40026        ' Wrong protocol for the requested transaction or request

  ' -- Errors specific to this class
  sckUnsupportedVersion = 40050   ' The version of Winsock is unsupported. Install Sockets version 2.0 or above.
  
  ' -- Windows Sockets specific errors
  sckInvalidArgument = 10014      ' The requested address is a broadcast address, but flag is not set.
  sckWouldBlock = 10035           ' Socket is non-blocking and the specified operation will block.
  sckInProgress = 10036           ' A blocking Winsock operation in progress.
  sckAlreadyComplete = 10037      ' The operation is completed. No blocking operation in progress
  sckNotSocket = 10038            ' The descriptor is not a socket.
  sckMsgTooBig = 10040            ' The datagram is too large to fit into the buffer and is truncated.
  sckPortNotSupported = 10043     ' The specified port is not supported.
  sckAddressInUse = 10048         ' Address in use.
  sckAddressNotAvailable = 10049  ' Address not available from the local machine.
  sckNetworkSubsystemFailed = 10050 ' Network subsystem failed.
  sckNetworkUnreachable = 10051   ' The network cannot be reached from this host at this time.
  sckNetReset = 10052             ' Connection has timed out when SO_KEEPALIVE is set.
  sckConnectAborted = 11053       ' Connection is aborted due to timeout or other failure.
  sckConnectionReset = 10054      ' The connection is reset by remote side.
  sckNoBufferSpace = 10055        ' No buffer space is available.
  sckAlreadyConnected = 10056     ' Socket is already connected.
  sckNotConnected = 10057         ' Socket is not connected.
  sckSocketShutdown = 10058       ' Socket has been shut down.
  sckTimedout = 10060             ' Socket has been shut down.
  sckConnectionRefused = 10061    ' Connection is forcefully rejected.
  sckNotInitialized = 10093       ' WinsockInit should be called first.
  sckHostNotFound = 11001         ' Authoritative answer: Host not found.
  sckHostNotFoundTryAgain = 11002 ' Non-Authoritative answer: Host not found.
  sckNonRecoverableError = 11003  ' Non-recoverable errors.
  sckNoData = 11004               ' Valid name, no data record of requested type.
End Enum

Private Enum EventIDs             ' Identifies an event for the RaiseEventEx routine.
  iCloseConnection = 1
  iConnect
  iConnectionRequest
  iDataArrival
  iError
  iSendComplete
  iSendProgress
End Enum

'===========================================================================
' Types
'===========================================================================
Private Type OVERLAPPED
  Internal            As Long
  InternalHigh        As Long
  Offset              As Long
  OffsetHigh          As Long
  hEvent              As Long
End Type

Private Type WNDCLASS
  Style               As Long
  lpfnwndproc         As Long
  cbClsextra          As Long
  cbWndExtra2         As Long
  hInstance           As Long
  hIcon               As Long
  hCursor             As Long
  hbrBackground       As Long
  lpszMenuName        As String
  lpszClassName       As String
End Type

Private Type POINTAPI
  x As Long
  y As Long
End Type

Private Type MSG
  hWnd As Long
  Message As Long
  wParam As Long
  lParam As Long
  time As Long
  pt As POINTAPI
End Type

'===========================================================================
'   Events (raised only if EventHandler is not implemented)
'===========================================================================
' NOTE: CloseConnection event is named Close in MSWINSCK.OCX
Public Event CloseConnection()
Public Event Connect()
Public Event ConnectionRequest(ByVal RequestID As Long)
Attribute ConnectionRequest.VB_Description = "Occurs when a remote client is attempting to connect."
Public Event DataArrival(ByVal BytesTotal As Long)
Attribute DataArrival.VB_Description = "Occurs when data has been received from the remote computer."
Attribute DataArrival.VB_MemberFlags = "200"
' NOTE: The signature on the Error event differs from MSWINSCK.OCX
Public Event Error(ByVal Number As Long, ByVal Description As String)
Public Event SendComplete()
Attribute SendComplete.VB_Description = "Occurs after a send operation has completed."
Public Event SendProgress(ByVal BytesSent As Long, ByVal BytesRemaining As Long)
Attribute SendProgress.VB_Description = "Occurs during process of sending data."

'===========================================================================
'   Private Members
'===========================================================================
Private m_hWnd                  As Long                 ' Window to receive Winsock events
Private m_WSAData               As WSAData              ' Socket version information
Private m_hSocket               As Long                 ' Handle to the socket
Private m_Protocol              As SocketProtocols      ' The protocol for this socket
Private m_RemoteAddr            As sockaddr_in          ' Connection info about remote machine [client]
Private m_RemoteHost            As String               '
Private m_LocalAddr             As sockaddr_in          ' Connection info about local  machine [server]
Private m_NetEvents             As FD_Events            ' Events selected to wait on
Private m_State                 As SocketStates         ' State of the socket
Private m_OutgoingData          As Collection           ' Outgoing asynchronous data
Private m_Aborted               As Boolean              ' Abort variable used when waiting on events
Private m_TextMode              As TextModes            ' Defines is text is sent/received as ANSI or Unicode
Private m_UserData              As Variant              ' Data owned by the caller

'===========================================================================
'   Public Members
'===========================================================================
Public EventHandler             As IWinsockEvents       ' Early bound event handler [great for server-side "listening" mode]
Attribute EventHandler.VB_VarDescription = "Pointer to a host that can receive early bound events via the IWinSoxEvents."

'===========================================================================
'   Declares
'===========================================================================
Private Declare Function GetProcessHeap Lib "kernel32" () As Long
Private Declare Function HeapAlloc Lib "kernel32" (ByVal hHeap As Long, ByVal dwFlags As Long, ByVal dwBytes As Long) As Long
Private Declare Function HeapDestroy Lib "kernel32" (ByVal hHeap As Long) As Long
Private Declare Function HeapFree Lib "kernel32" (ByVal hHeap As Long, ByVal dwFlags As Long, ByVal lpMem As Long) As Long
Private Declare Function HeapReAlloc Lib "kernel32" (ByVal hHeap As Long, ByVal dwFlags As Long, ByVal lpMem As Long, ByVal dwBytes As Long) As Long
Private Declare Function HeapSize Lib "kernel32" (ByVal hHeap As Long, ByVal dwFlags As Long, ByVal lpMem As Long) As Long
Private Declare Function CreateWindowEx Lib "user32" Alias "CreateWindowExA" (ByVal dwExStyle As Long, ByVal lpClassName As String, ByVal lpWindowName As String, ByVal dwStyle As Long, ByVal x As Long, ByVal y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal hWndParent As Long, ByVal hMenu As Long, ByVal hInstance As Long, lpParam As Any) As Long
Private Declare Function DestroyWindow Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function IsWindow Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function RegisterClass Lib "user32" Alias "RegisterClassA" (Class As WNDCLASS) As Long
Private Declare Function RegisterWindowMessage Lib "user32" Alias "RegisterWindowMessageA" (ByVal lpString As String) As Long
Private Declare Function GetProp Lib "user32" Alias "GetPropA" (ByVal hWnd As Long, ByVal lpString As String) As Long
Private Declare Function SetProp Lib "user32" Alias "SetPropA" (ByVal hWnd As Long, ByVal lpString As String, ByVal hData As Long) As Long
Private Declare Function RemoveProp Lib "user32" Alias "RemovePropA" (ByVal hWnd As Long, ByVal lpString As String) As Long
Private Declare Sub OutputDebugStringA Lib "kernel32" (ByVal lpOutputString As String)

Private Declare Function PeekMessage Lib "user32" Alias "PeekMessageA" (lpMsg As MSG, ByVal hWnd As Long, ByVal wMsgFilterMin As Long, ByVal wMsgFilterMax As Long, ByVal wRemoveMsg As Long) As Long
Private Declare Function GetMessage Lib "user32" Alias "GetMessageA" (lpMsg As MSG, ByVal hWnd As Long, ByVal wMsgFilterMin As Long, ByVal wMsgFilterMax As Long) As Long
Private Declare Function TranslateMessage Lib "user32" (lpMsg As MSG) As Long
Private Declare Function DispatchMessage Lib "user32" Alias "DispatchMessageA" (lpMsg As MSG) As Long

Public Sub Accept(ByVal RequestID As Long)
Attribute Accept.VB_Description = "Accept an incoming connection request."
'===========================================================================
' Accept - Accepts an incoming connection via the ConnectionRequest event.
'===========================================================================

  Dim hSocket     As Long
  Dim ErrNum      As Long
  Dim ErrMsg      As String
  
  ' Throw error if this is not a fresh/closed socket.
  If Me.State <> sckClosed And Me.State <> sckListening Then ErrRaise sckBadState
  If Me.Protocol <> sckTCPProtocol Then ErrRaise sckBadState
  
  On Error GoTo ErrHandler

  ' Wait for the client connection
  hSocket = MWinsock.Accept(RequestID, m_RemoteAddr, Len(m_RemoteAddr))
  If hSocket = INVALID_SOCKET Then
    WsaRaiseError Err.LastDllError, Module
  End If
  
  ' If the caller "Accepts" on the listening socket then
  ' we want to make sure to shut the listening socket down.
  Call DestroySocket
  
  ' Save the new socket we have just Accept'ed
  m_hSocket = hSocket
  
  ' Create window to receive socket-based messages
  Call CreatePrivateWindow

  ' Create a list to hold outbound write requests.
  Set m_OutgoingData = New Collection

  ' Attach the event to the socket. Save net events for subsequent calls.
  m_NetEvents = FD_READ Or FD_WRITE Or FD_CLOSE
  If WSAAsyncSelect(m_hSocket, m_hWnd, SocketMessage(), m_NetEvents) = SOCKET_ERROR Then
    WsaRaiseError Err.LastDllError, Module
  End If
  
  ' Notify the host and allow them to call the
  ' Accept method. Ignore errors.
  State = sckConnected

  ' Notify owner we are connected
  RaiseEventEx iConnect

ExitLabel:
  ' -- Exit and return control back to the caller. --
  On Error Resume Next
  If ErrNum <> 0 Then
    ' Tear down the socket and re-raise the error
    Call DestroySocket
    On Error GoTo 0
    ErrRaise ErrNum, ErrMsg
  End If

  Exit Sub
  Resume
ErrHandler:
  ' Save error info
  ErrNum = Err.Number: ErrMsg = Err.Description
  Debug.Assert 0
  Resume ExitLabel
    
End Sub

Public Sub Bind( _
  Optional ByVal LocalPort As Long, _
  Optional ByVal LocalIP As String)
'===========================================================================
' Bind - Allows the caller to bind to a specific port.
'===========================================================================

  Dim addr        As sockaddr_in
  Dim ErrNum      As Long
  Dim ErrMsg      As String
  
  ' Throw error if this is not a fresh/closed socket.
  If Me.State <> sckClosed Then ErrRaise sckBadState

  On Error GoTo ErrHandler
  
  ' Connect to the desired port
  m_hSocket = GetSocket()
  
  ' Update our current state
  State = sckOpen

  ' Make copy of local addr info since we might
  ' overload the the port and IP address.
  addr = m_LocalAddr
  With addr
      
    ' We allow the caller to specify the LocalIP since there
    ' could be multiple NIC's on the box.
    If Len(LocalIP) > 0 Then
      ' Convert IP address string to an internet address structure
      .sin_addr.s_addr = inet_addr(LocalIP)
    ElseIf .sin_addr.s_addr <= 0 Then
      ' Use any interface if LocalHost not set
      .sin_addr.s_addr = ADDR_ANY
    End If
    
    ' Bind to a specific port if supplied
    ' (overridding the LocalPort property).
    If LocalPort > 0 Then
      ' Convert to network byte order
      .sin_port = htons(LocalPort)
    End If
      
  End With
  
  Debug.Assert Len(m_LocalAddr) = 16
  ' Attach the socket to the IP address
  If MWinsock.Bind(m_hSocket, addr, Len(addr)) = SOCKET_ERROR Then
    WsaRaiseError Err.LastDllError, Module
  End If

ExitLabel:
  ' -- Exit and return control back to the caller. --
  On Error Resume Next
  If ErrNum <> 0 Then
    On Error GoTo 0
    ErrRaise ErrNum, ErrMsg
  End If

  Exit Sub
  Resume
ErrHandler:
  ' Save error info
  ErrNum = Err.Number: ErrMsg = Err.Description
  Debug.Assert 0
  Resume ExitLabel
    
End Sub

' NOTE: CloseConnection property is named Close in MSWINSCK.OCX
Public Sub CloseConnection()
Attribute CloseConnection.VB_Description = "Close current connection."
'===========================================================================
' CloseConnection - Destroys the sockets.
'===========================================================================

  If Me.State <> sckClosed Then

    Call DestroySocket

    On Error Resume Next
    RaiseEventEx iCloseConnection

  End If

End Sub

Public Sub Connect( _
  Optional ByVal RemoteHost As String, _
  Optional ByVal RemotePort As Long, _
  Optional ByVal Synchronous As Boolean = False)
'===========================================================================
' Connect - Connects to another local/remote socket via TCP.
'===========================================================================

  Dim Host            As hostent
  Dim lpHost          As Long
  Dim ErrorCode       As SocketErrors
  Dim ErrNum          As Long
  Dim ErrMsg          As String

  If Me.State <> sckClosed Then ErrRaise sckBadState

  On Error GoTo ErrHandler

  ' Use property values if no host/port info supplied
  If Len(RemoteHost) = 0 Then RemoteHost = Me.RemoteHost
  If RemotePort <= 0 Then RemotePort = Me.RemotePort

  ' Connect to the desired port
  m_hSocket = GetSocket()
  
  With m_RemoteAddr
  
    ' Specify IP-based protocol
    .sin_family = AF_INET
    
    ' Convert port to network byte order
    .sin_port = htons(RemotePort)
    
    ' Convert IP address string or hostname
    ' to an internet address structure
    State = sckResolvingHost
    .sin_addr.s_addr = AnyToAddress(RemoteHost)
    State = sckHostResolved
      
  End With

  ' NOTE: We want to call the connect() function before we return
  ' control back to the caller in case connect() results in an
  ' error of anything other than WSAEWOULDBLOCK, in which case we
  ' could error out in the method, rather than raising an error event.

  If Not Synchronous Then
    
    ' WSAAsyncSelect is supposed to set the socket in non-blocking mode
    ErrorCode = ioctlsocket(m_hSocket, FIONBIO, 1&)
    Debug.Assert (ErrorCode <> SOCKET_ERROR)
    
    ' Attach the event to the socket. Save net events for subsequent calls.
    m_NetEvents = FD_READ Or FD_CONNECT Or FD_CLOSE
    If WSAAsyncSelect(m_hSocket, m_hWnd, SocketMessage(), m_NetEvents) = SOCKET_ERROR Then
      WsaRaiseError Err.LastDllError, Module
    End If
  
  End If

  ' Update state variable
  State = sckConnecting
  
  ' Begin the connection process to the server
  'ErrorCode = MWinsock.Connect(m_hSocket, m_RemoteAddr, Len(m_RemoteAddr))
  ErrorCode = WSAConnect( _
    m_hSocket, _
    m_RemoteAddr, Len(m_RemoteAddr), _
    ByVal 0, ByVal 0, ByVal 0, ByVal 0)
    
  If ErrorCode = SOCKET_ERROR Then
    If Err.LastDllError <> WSAEWOULDBLOCK Then
      WsaRaiseError Err.LastDllError, Module
    End If
  Else
    ' We have synchronously connected successfully
    State = sckConnected
  End If
  
  ' Create a list to hold outbound write requests.
  Set m_OutgoingData = New Collection
  
  If Synchronous Then

    ' Our synchronous connection was a success.
    ' Begin looking for read/close events
    m_NetEvents = FD_READ Or FD_CLOSE
    If WSAAsyncSelect(m_hSocket, m_hWnd, SocketMessage(), m_NetEvents) = SOCKET_ERROR Then
      WsaRaiseError Err.LastDllError, Module
    End If

  End If
  
ExitLabel:
  ' -- Exit and return control back to the caller. --
  On Error Resume Next
  If ErrNum <> 0 Then
    ' Tear down the socket and re-raise the error
    Call DestroySocket
    On Error GoTo 0
    ErrRaise ErrNum, ErrMsg
  End If

  Exit Sub
  Resume
ErrHandler:
  ' Save error info
  ErrNum = Err.Number: ErrMsg = Err.Description
  Debug.Assert 0
  Resume ExitLabel

End Sub
 
Public Sub GetData( _
  Data, _
  Optional ByVal DataType As VbVarType, _
  Optional ByVal MaxLen As Long)
'===========================================================================
' GetData - Reads data from the socket buffer. The contents read will be
' purged from the buffer.
'===========================================================================

  ' Throw error if socket not connected
  If Me.State <> sckConnected And Me.State <> sckListening Then ErrRaise sckBadState

  ' Fetch data from the socket
  DataRecv Data, DataType, m_hSocket, VarPtr(m_Aborted), MaxLen, 0

End Sub

Public Function GetDataAsString(Optional ByVal MaxLen As Long) As String
'===========================================================================
' GetDataAsString - Reads data from the socket buffer and returns it as a
' string.
'===========================================================================

  Dim sData           As String
  
  ' Throw error if socket not connected
  If Me.State <> sckConnected And Me.State <> sckListening Then ErrRaise sckBadState

  ' Fetch data from the socket
  GetData sData, vbString
  
  ' Return the data
  GetDataAsString = sData

End Function

Public Sub Listen() 'Optional ByVal LocalPort As Long, Optional ByVal LocalIP As String)
Attribute Listen.VB_Description = "Listen for incoming connection requests."
'===========================================================================
' Listen - Listens for incoming TCP connections on a specific port and on an
' optional NIC.
'===========================================================================
'Creates a socket and sets it in listen mode. This method works only for TCP connections.

  Dim ErrorCode       As Long
  Dim ErrNum          As Long
  Dim ErrMsg          As String
  Dim hClientSocket   As Long
  
  ' Throw error if socket not closed
  If Me.State <> sckClosed And Me.State <> sckOpen Then ErrRaise sckBadState

  On Error GoTo ErrHandler

  ' If not bound, perform the bind'ing automatically.
  If Me.State <> sckOpen Then
    Me.Bind 'LocalPort, LocalIP
  End If
  
  ' Begin listening on the local port waiting for connections
  ErrorCode = MWinsock.Listen(m_hSocket, SOMAXCONN)
  If ErrorCode = SOCKET_ERROR Then
    If Err.LastDllError <> WSAEWOULDBLOCK Then
      WsaRaiseError Err.LastDllError, Module
    End If
  End If

  ' Create a list to hold outbound write requests.
  Set m_OutgoingData = New Collection

  ' Attach the event to the socket. Save net events for subsequent calls.
  m_NetEvents = FD_ACCEPT Or FD_CLOSE
  If WSAAsyncSelect(m_hSocket, m_hWnd, SocketMessage(), m_NetEvents) = SOCKET_ERROR Then
      WsaRaiseError Err.LastDllError, Module
  End If

  ' Update the state variable
  State = sckListening

ExitLabel:
  ' -- Exit and return control back to the caller. --
  On Error Resume Next
  If ErrNum <> 0 Then
    ' Tear down the socket and re-raise the error
    Call DestroySocket
    On Error GoTo 0
    ErrRaise ErrNum, ErrMsg
  End If

  Exit Sub
  Resume
ErrHandler:
  ' Save error info
  ErrNum = Err.Number: ErrMsg = Err.Description
  Debug.Assert 0
  Resume ExitLabel
    
End Sub

Public Sub PeekData( _
  Data, _
  Optional ByVal DataType As VbVarType, _
  Optional ByVal MaxLen As Long)
'===========================================================================
' PeekData - Reads data without purging it from the socket's data buffer.
'===========================================================================

  ' Throw error if socket not connected
  Select Case Me.State
  Case sckConnected, sckListening
    ' Valid states; do nothing
  Case Else
    ErrRaise sckBadState
  End Select
  
  ' Fetch data from the socket
  DataRecv Data, DataType, m_hSocket, VarPtr(m_Aborted), MaxLen, MSG_PEEK

End Sub

Public Function PeekDataAsString(Optional ByVal MaxLen As Long) As String
'===========================================================================
' PeekDataAsString - Reads data from the socket buffer and returns it as a
' string, without purging it from the socket's data buffer.
'===========================================================================

  Dim sData           As String
  
  ' Throw error if socket not connected
  If Me.State <> sckConnected And Me.State <> sckListening Then ErrRaise sckBadState

  ' Fetch data from the socket
  PeekData sData, vbString
  
  ' Return the data
  PeekDataAsString = sData

End Function

Public Sub SendData( _
  Data, _
  Optional ByVal DataLen As Long = -1, _
  Optional ByVal Synchronous As Boolean = False)
'===========================================================================
' SendData - Sends string or binary data to another socket.
'
' Data            A Variant of subtype string or byte array, or a long pointer
'                 to the data to send.
' DataLen         Optional. Required when data supplied is of type long
'                 (assumed to be a pointer).
' Synchronous     Optional. True to send the data synchronously. If True the
'                 call will not return until the write completes or times out.
'                 Progress and completion events will be raised as usual.
'===========================================================================

  Dim cbData          As Long
  Dim WorkString      As String
  Dim lpData          As Long
  Dim lpBuffer        As Long
  Dim ErrNum          As Long
  Dim ErrMsg          As String

  ' Throw error if socket not connected
  On Error GoTo 0
  If Me.State <> sckConnected And Me.State <> sckListening Then ErrRaise sckBadState

  On Error GoTo ErrHandler

  If Synchronous Then

    DataSend Data, DataLen, True
    DoSocketEvents
    GoTo ExitLabel

  End If

  If VarType(Data) = (vbArray Or vbByte) Then
    ' A byte array was supplied. No conversion needed.

    Dim sa      As SAFEARRAY

    ' Throw error if invalid/empty array supplied
    If Not GetSafeArrayInfo(Data, sa) Then Call ApiRaise(Err.LastDllError, Module)
    If sa.cbElements <= 0 Then ErrRaise 5
    If sa.cDims > 1 Then ErrRaise 5

    cbData = sa.rgsabound(1).cElements
    lpData = VarPtr(Data(LBound(Data))) 'sa.pvData

  ElseIf VarType(Data) = vbLong Then

    lpData = Data
    cbData = DataLen

  ElseIf VarType(Data) = vbString Then
  
    If Me.TextMode = tmUnicode Then
      ' Leave as is -- send as Unicode
      lpData = StrPtr(Data)
      cbData = Len(Data) * 2
    Else
      ' Convert to Ansi before sending
      WorkString = StrConv(Data, vbFromUnicode)
      lpData = StrPtr(WorkString)
      cbData = Len(Data)
    End If

  Else

    ErrRaise 13 ' Type Mismatch
  
  End If

  ' Create a storage buffer and append it to the queue.
  lpBuffer = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY Or HEAP_NO_SERIALIZE, cbData)
  CopyMemory ByVal lpBuffer, ByVal lpData, cbData
  m_OutgoingData.Add lpBuffer

  ' Attach the event to the socket. Save net events for subsequent calls.
  m_NetEvents = m_NetEvents Or FD_WRITE

  If WSAAsyncSelect(m_hSocket, m_hWnd, SocketMessage(), m_NetEvents) = SOCKET_ERROR Then
    WsaRaiseError Err.LastDllError, Module
  End If

ExitLabel:
  ' -- Exit and return control back to the caller. --
  On Error Resume Next
  If ErrNum <> 0 Then
    On Error GoTo 0
    ErrRaise ErrNum, ErrMsg
  End If

  Exit Sub
  Resume
ErrHandler:
  ' Save error info
  ErrNum = Err.Number: ErrMsg = Err.Description
  Debug.Assert 0
  Resume ExitLabel
    
End Sub


Public Property Get BytesReceived() As Long
Attribute BytesReceived.VB_Description = "Returns the number of bytes that can be read from the socket (which 'may' differ than the number of bytes actually received)."
'===========================================================================
' BytesReceived - Returns the number of bytes in the socket data buffer.
'===========================================================================

  Dim ErrorCode       As SocketErrors
  
  ' Throw error if socket not connected
  If Me.State <> sckConnected And Me.State <> sckListening Then ErrRaise sckBadState
  
  ' Query largest size available for a call to recv
  ErrorCode = ioctlsocket(m_hSocket, FIONREAD, BytesReceived)
  If ErrorCode = SOCKET_ERROR Then
    WsaRaiseError Err.LastDllError, Module
  End If

End Property

Public Property Get LocalHostName() As String
Attribute LocalHostName.VB_Description = "Returns the local machine name."
'===========================================================================
' LocalHostName - Returns the name of the "localhost"
'===========================================================================

  Dim HostName        As String
  Dim cbHostName      As Long
  
  ' Initialize output buffer
  cbHostName = 256
  HostName = String(cbHostName, 0)
  
  ' Fetch local host
  If gethostname(HostName, cbHostName) = SOCKET_ERROR Then
    WsaRaiseError Err.LastDllError, Module
  End If
  
  ' Trim off nulls
  LocalHostName = TrimNulls(HostName)
    
End Property

Public Property Get LocalIP() As String
Attribute LocalIP.VB_Description = "Returns the local machine IP address."
'===========================================================================
' LocalHostIP - Returns the IP address of the "localhost"
'===========================================================================
  LocalIP = AddressToIP(NameToAddress(LocalHostName))
  'LocalIP = AddressToIP(m_LocalAddr.sin_addr.s_addr)
End Property

Public Property Get LocalPort() As Long
Attribute LocalPort.VB_Description = "Returns/Sets the port used on the local computer."
'===========================================================================
' Returns the local port on which the socket is connected or listening.
'===========================================================================

  ' Convert to network byte order
  LocalPort = ntohs(m_LocalAddr.sin_port)

End Property
Public Property Let LocalPort(ByVal NewValue As Long)

  ' Throw error if the socket is bound or connected
  If Me.State <> sckClosed Then ErrRaise sckBadState

  m_LocalAddr.sin_port = htons(NewValue)

End Property

Public Property Get Protocol() As SocketProtocols
Attribute Protocol.VB_Description = "Returns/Sets the socket protocol."
'===========================================================================
' Protocol - Allows the caller to choose between UDP and TCP protocols.
'===========================================================================

  Protocol = m_Protocol
  
End Property
Public Property Let Protocol(ByVal NewValue As SocketProtocols)

  ' Throw error if this is not a fresh/closed socket.
  If Me.State <> sckClosed Then ErrRaise sckBadState
  
  Select Case NewValue
  Case sckUDPProtocol, sckTCPProtocol
    ' Specify IP-based protocol
    m_LocalAddr.sin_family = AF_INET

' TODO:
' NOT supported yet:
' Case sckIrDA
'
'   ' Specify IrDA protocol
'   m_LocalAddr.sin_family = AF_IRDA

  Case Else
      ErrRaise 380
  End Select
  
  m_Protocol = NewValue
    
End Property

Public Property Get RemoteHost() As String
Attribute RemoteHost.VB_Description = "Returns/Sets the name used to identify the remote computer."
Attribute RemoteHost.VB_UserMemId = 0
'===========================================================================
' RemoteHost - Returns the name of the remote machine.
'===========================================================================

  RemoteHost = m_RemoteHost
  If Len(m_RemoteHost) = 0 Then
    RemoteHost = AddressToName(m_RemoteAddr.sin_addr.s_addr)
  End If
  
End Property
Public Property Let RemoteHost(ByVal NewValue As String)

  ' TODO: Verify logic for this test with UDP
  If (Me.Protocol = sckTCPProtocol) Then
    If (Me.State <> sckClosed) Then
      ErrRaise sckBadState
    End If
'  Or (Me.Protocol = sckUDPProtocol And Me.State = sckListening) Then
'    ErrRaise sckBadState
  End If

  m_RemoteAddr.sin_addr.s_addr = NameToAddress(NewValue)
  m_RemoteHost = NewValue
  
End Property

Public Property Get RemoteHostIP() As String
Attribute RemoteHostIP.VB_Description = "Returns the remote host IP address."
'===========================================================================
' RemoteHost - Returns the IP address of the remote machine.
'===========================================================================

  RemoteHostIP = AddressToIP(m_RemoteAddr.sin_addr.s_addr)

End Property

Public Property Get RemotePort() As Long
Attribute RemotePort.VB_Description = "Returns/Sets the port to be connected to on the remote computer."
'===========================================================================
' RemotePort - Returns the port on which the remote socket is connected/bound
' to.
'===========================================================================

  RemotePort = ntohs(m_RemoteAddr.sin_port)

End Property
Public Property Let RemotePort(ByVal NewValue As Long)

  Select Case Me.Protocol
  Case sckTCPProtocol
    If Me.State <> sckClosed Then
      ErrRaise sckBadState
    End If
  Case sckUDPProtocol
    'If Me.State <> sckClosed Then
    ' TODO: I don't know if this is legal if connect()
    ' was called on the UDP socket.
  Case Else
    Debug.Assert 0
    ErrRaise 5
  End Select
  
  m_RemoteAddr.sin_port = htons(NewValue)

End Property

Public Property Get SocketHandle() As Long
Attribute SocketHandle.VB_Description = "Returns the socket handle."
'===========================================================================
' SocketHandle - Returns handle of the socket. This allows the caller to
' perhaps clone the socket or perform some other unsupported operation.
'===========================================================================
  SocketHandle = m_hSocket
End Property

Public Property Get State() As SocketStates
Attribute State.VB_Description = "Returns the state of the socket connection."
'===========================================================================
' State - Returns a value representing the current state of the socket.
'===========================================================================
  State = m_State
End Property
Private Property Let State(ByVal NewValue As SocketStates)

  ' Change to new state
  m_State = NewValue
  
  Select Case m_State
  Case sckClosed, sckClosing
    ' Setting this value will kill internal "WaitXxx" calls
    m_Aborted = True
  Case Else
    m_Aborted = False
  End Select
    
End Property


Public Property Get TextMode() As TextModes
Attribute TextMode.VB_Description = "Defines if data of type String is sent/received as Ansi or Unicode."
'===========================================================================
' TextMode - Defines if text is sent/received as Ansi or Unicode. The Value
' is set to tmAnsi by default (text is converted to/from Ansi).
'===========================================================================
  TextMode = m_TextMode
End Property
Public Property Let TextMode(ByVal NewValue As TextModes)
  Select Case NewValue
  Case tmAnsi, tmUnicode
    m_TextMode = NewValue
  Case Else
    ErrRaise 380
  End Select
End Property


Public Property Get UserData() As Variant
'===========================================================================
' UserData - User defined data. Not used by the class.
'===========================================================================
  If IsObject(m_UserData) Then
    Set UserData = m_UserData
  Else
    UserData = m_UserData
  End If
End Property
Public Property Let UserData(ByVal NewValue As Variant)
  m_UserData = NewValue
End Property
Public Property Set UserData(ByVal NewRef As Variant)
  Set m_UserData = NewRef
End Property



Public Property Get SendTimeout() As Long
'===========================================================================
' SendTimeout - Optional property. Defines the # of seconds to time out
' on remote calls.
'===========================================================================
  Dim Failure As Long
  
  If InvalidHandleValue(m_hSocket) Then Exit Property

  ' Get the timout value
  Failure = getsockopt(m_hSocket, SOL_SOCKET, SO_SNDTIMEO, _
    ByVal VarPtr(SendTimeout), Len(SendTimeout))
  
  ' Throw on failure
  If Failure <> 0 Then
    Err.Raise Err.LastDllError
  End If
  
End Property
Public Property Let SendTimeout(ByVal NewValue As Long)
  
  Dim Failure As Long

  If InvalidHandleValue(m_hSocket) Then ErrRaise sckBadState
  If NewValue < 0 Then Err.Raise 380 ' Invalid Property Value

  ' Make the change
  Failure = setsockopt(m_hSocket, SOL_SOCKET, SO_SNDTIMEO, _
    ByVal VarPtr(NewValue), Len(NewValue))
  
  ' Throw on failure
  If Failure <> 0 Then
    Err.Raise Err.LastDllError
  End If
  
End Property


Public Property Get ReceiveTimeout() As Long
'===========================================================================
' ReceiveTimeout - Optional property. Defines the # of seconds to time out
' on remote calls.
'===========================================================================
  
  Dim Failure As Long
  
  If InvalidHandleValue(m_hSocket) Then Exit Property

  ' Get the timout value
  Failure = getsockopt(m_hSocket, SOL_SOCKET, SO_RCVTIMEO, _
    ByVal VarPtr(ReceiveTimeout), Len(ReceiveTimeout))
  
  ' Throw on failure
  If Failure <> 0 Then
    Err.Raise Err.LastDllError
  End If
  
End Property
Public Property Let ReceiveTimeout(ByVal NewValue As Long)

  Dim Failure As Long
  
  If InvalidHandleValue(m_hSocket) Then ErrRaise sckBadState
  If NewValue < 0 Then Err.Raise 380 ' Invalid Property Value

  ' Make the change
  Failure = setsockopt(m_hSocket, SOL_SOCKET, SO_RCVTIMEO, _
    ByVal VarPtr(NewValue), Len(NewValue))
  
  ' Throw on failure
  If Failure <> 0 Then
    Err.Raise Err.LastDllError
  End If
  
End Property



'===========================================================================
' Private Functions
'===========================================================================


Private Sub Destroy()
'===========================================================================
' Destroy - Tears down the class and the socket library. NOTE: The socket
' library is only torn down if there are no other open class references.
'===========================================================================

  Dim ErrorCode   As SocketErrors
  Dim MemObject   As Variant
  
  On Error Resume Next
  
  ' Kill any event waits
  m_Aborted = True
  DoSocketEvents

  ' Destroy the socket
  Call DestroySocket
  
  If ErrorCode <> NO_ERROR Then
    Debug.Print "CloseSocket=" & ErrorCode
  End If

  ' Kill our callback window
  If Not InvalidHandleValue(m_hWnd) Then
    RemoveProp m_hWnd, App.EXEName
    DestroyWindow m_hWnd
    m_hWnd = 0
  End If

  ' Free memory objects used to receive data into, if any
  For Each MemObject In m_OutgoingData
    HeapFree GetProcessHeap(), HEAP_NO_SERIALIZE, MemObject
  Next MemObject
  Set m_OutgoingData = Nothing
    
End Sub


Private Sub DataRecv( _
  Data, _
  ByVal DataType As VbVarType, _
  ByVal hSocket As Long, _
  ByVal lpAbortVariable As Long, _
  Optional ByVal MaxLen As Long, _
  Optional RecvFlags As MessageFlags)
'===========================================================================
' DataRecv - Internal call to read data from the socket read buffer.
'===========================================================================

  Dim ErrorCode       As SocketErrors
  Dim Success         As Boolean
  Dim WorkArray()     As Byte                 ' Used only when input data type is vbString
  Dim cbReceived      As Long                 ' Sum of bytes read
  Dim lpBytesIn       As Long                 ' # of bytes read for each iteration
  Dim Flags           As Long                 ' WSARecv in/out flags [like MSG_PEEK]
  Dim OL              As OVERLAPPED           ' Needed for overlapped reads.
  Dim Bfr             As WSABUF               ' outbound buffer [scatter supported]
  Dim ErrNum          As Long
  Dim ErrMsg          As String

  ' Identify how large the buffer is.
  cbReceived = Me.BytesReceived
  
  ' Validate MaxLen
  Select Case MaxLen
  Case Is < 0
    ErrRaise 5
  Case 0, Is > cbReceived
    If VarType(Data) = vbLong Then ErrRaise 5 ' todo
    MaxLen = cbReceived
  Case Is < cbReceived, cbReceived
    '
  End Select

  ' Validate DataType
  Select Case DataType
  Case vbString, (vbByte Or vbArray)
    ' this is cool
  Case Else
    DataType = VarType(Data)
  End Select
  
  
  ' Prepare input buffer
  Select Case DataType
  Case vbString
    ' We will be returning the data as a string.
    ' We use a byte array instead of the input string
    ' variable since the data may not be divisible by
    ' two, and BSTR's are 2 bytes per character.
    
    ' Handle case of empty packet
    If MaxLen <= 0 Then
      Data = ""
      GoTo ExitLabel
    End If
    
    ' Size buffer to fit
    ReDim WorkArray(0 To MaxLen - 1)
    ' Point to the buffer for recv
    Bfr.Buf = VarPtr(WorkArray(0))

  Case (vbArray Or vbByte)
    ' We will be returning the data as a byte array
    
    ' Handle case of empty packet
    If MaxLen <= 0 Then
      Erase Data
      GoTo ExitLabel
    End If
    
    ' Size buffer to fit
    ReDim Data(0 To MaxLen - 1)
    
    ' Point to the buffer for recv
    Bfr.Buf = VarPtr(Data(0))
      
  Case vbLong
    ' Data is assumed to be a long pointer
    
    ' Error if no length specified or appears to be a bogus handle.
    If MaxLen <= 0 Or InvalidHandleValue(Data) Then
        Err.Raise 5
    End If
    ' Point to the buffer for recv
    Bfr.Buf = Data
      
  Case Else
    ErrRaise 5
  End Select
  
  ' Size of the buffer
  Bfr.len = MaxLen

  ' -- Prepare to read the data --
  
  ' Create the event to wait on
  OL.hEvent = WSACreateEvent()
  If InvalidHandleValue(OL.hEvent) Then
    ' Something went wrong. Notify host but ignore their errors.
    On Error Resume Next
    RaiseEventEx iError, Err.LastDllError Or vbObjectError, _
        "Could not create write event. " & LastDllErrorMsg(Err.LastDllError)
    On Error GoTo ErrHandler
  End If
  
  ' Begin reading from the read buffer.
  ' Loop until everything read or we error out.
  cbReceived = 0
  Do While (cbReceived < Bfr.len)

    ' Clear output variables
    Flags = RecvFlags
    lpBytesIn = 0
    
    ' Fill buffer with incoming data.
    ErrorCode = WSARecv(hSocket, Bfr, 1, _
      lpBytesIn, Flags, OL, ByVal 0&)
    If ErrorCode = SOCKET_ERROR Then
      ErrorCode = Err.LastDllError
    End If
    
    ' ** Flags ignored. MSG_PARTIAL not supported **

    Select Case ErrorCode
    Case NO_ERROR

      ' data received without waiting
      cbReceived = cbReceived + lpBytesIn
        
    Case ERROR_IO_PENDING

      ' Need to wait on our event
      If WaitOnEvent(OL.hEvent, Me.ReceiveTimeout, lpAbort:=lpAbortVariable) > 0 Then
          ' Data is available

        ' How many bytes were received?
        Success = WSAGetOverlappedResult(hSocket, OL, lpBytesIn, 1, Flags)
        If Not Success Then
          ' Something went wrong. Throw an error.
          WsaRaiseError Err.LastDllError, Module, "Could not read data."
        End If
        
        ' Keep track of how much data is in the read buffer area.
        cbReceived = cbReceived + lpBytesIn
        If cbReceived = 0 Then
          Debug.Assert 0
          ' Should we throw an error to prevent potential loop?
        End If
          
      Else

        ' Something went wrong. Throw an error.
        WsaRaiseError Err.LastDllError, Module, "Could not read data."
          
      End If

      ' Reset the event for the next pass.
      WSAResetEvent OL.hEvent

    Case Else

      ' Something went wrong. Throw an error.
      WsaRaiseError Err.LastDllError, Module, "Could not read data."

    End Select
      
  Loop

  ' If fetching string data, and data is sent as
  ' Ansi (the default) then it is converted to Unicode
  ' before it is returned to the caller.
  If DataType = vbString Then
    If Me.TextMode = tmAnsi Then
      ' Convert to Unicode
      Data = StrConv(WorkArray, vbUnicode)
    Else
      ' Just move the byte array to the string variable.
      Data = WorkArray
    End If
  End If

ExitLabel:
  ' -- Exit and return control back to the caller. --
  On Error Resume Next
  WSACloseEvent OL.hEvent
  If ErrNum <> 0 Then
    On Error GoTo 0
    ErrRaise ErrNum, ErrMsg
  End If

  Exit Sub
  Resume
ErrHandler:
  ' Save error info
  ErrNum = Err.Number: ErrMsg = Err.Description
  Debug.Assert 0
  Resume ExitLabel

End Sub


Private Sub DataSend(Data, Optional ByVal DataLen As Long, _
                     Optional ByVal Synchronous As Boolean = False)
'===========================================================================
' DataSend - Internal call to send binary or string data to another socket.
' This call behaves differently between synchronous/asynchronous modes.
'===========================================================================

  Dim ErrorCode       As SocketErrors     ' Return values
  Dim cbDataSent      As Long             ' OUT. # of bytes sent by WSASend
  Dim cbTotalSent     As Long             ' The sum of bytes sent via WSASend
  Dim Bfr             As WSABUF           ' outbound buffer [scatter not supported]
  Dim WorkString      As String           ' Used only when sending "Ansi" text.
  Dim Flags           As Long             ' Misc overlap flags
  Dim OL              As OVERLAPPED       ' Stores event for overlapped I/O
  Dim ErrNum          As Long
  Dim ErrMsg          As String
  Dim i               As Long

  ' Trap ALL errors.
  On Error GoTo ErrHandler

  ' Fill the output buffer to be used to send data
  With Bfr

    If VarType(Data) = vbLong Then
      ' Internal calls only.
      ' Is a pointer to data stored in the internal
      ' outbound queue -- asynchronous write operations only.

' TODO: Monte, Looks like you added support for
' external _synchronous_ writes. Need to test that it works.

      ' Error if not an an async write (caller supplied wrong data type).
      ' This data type reserved for internal calls only.
      If Synchronous Then 'ErrRaise 13
          
        ' Save size of buffer stored on the heap
        .len = DataLen
      
      Else
        ' Save size of buffer stored on the heap
        .len = HeapSize(GetProcessHeap(), 0&, Data)

      End If
      
      ' Save pointer to array's data
      .Buf = Data

    ElseIf VarType(Data) = (vbArray Or vbByte) Then
      ' A byte array was supplied. No conversion needed.
      
      ' Fetch info about the supplied array
      Dim sa      As SAFEARRAY

      ' Throw error if invalid/empty array supplied
      If Not GetSafeArrayInfo(Data, sa) Then Call ApiRaise(Err.LastDllError, Module)
      If sa.cbElements <= 0 Then ErrRaise 5
      If sa.cDims > 1 Then ErrRaise 5

      ' Save pointer to array's data
      '.Buf = sa.pvData
      .Buf = VarPtr(Data(LBound(Data)))
      ' Save # of elements in byte array
      .len = sa.rgsabound(1).cElements

    ElseIf VarType(Data) = vbString Then
      ' A string was supplied. Convert to ANSI if needed.
      
      If Me.TextMode = tmUnicode Then
        ' Leave as is -- send as Unicode
        .Buf = StrPtr(Data)
        .len = Len(Data) * 2
      Else
        ' Convert to Ansi before sending
        WorkString = StrConv(Data, vbFromUnicode)
        .Buf = StrPtr(WorkString)
        .len = Len(Data)
      End If
    
    Else
      ' Unsupported data type supplied
      ErrRaise 13 ' Type Mismatch
        
    End If

  End With
  
  ' Create the event to wait on
  OL.hEvent = WSACreateEvent()
  If InvalidHandleValue(OL.hEvent) Then
    ApiRaise Err.LastDllError, Module, "Could not create write event."
  End If

  If Me.Protocol = sckUDPProtocol Then
    ' Initialize addr_in structure outside of loop
    With m_RemoteAddr
      .sin_family = AF_INET
      .sin_addr.s_addr = AnyToAddress(m_RemoteHost)
    End With
  End If
  
  ' Initialize. Allow for zero Bfr.Len in case
  ' an empty packet is being sent.
  cbTotalSent = 0
  Do

    ' Write output buffer(s) to the socket.
    ErrorCode = NO_ERROR
    
    If Me.Protocol = sckTCPProtocol Then
      ErrorCode = WSASend(m_hSocket, ByVal VarPtr(Bfr), 1&, cbDataSent, _
        Flags, ByVal VarPtr(OL), 0&)
    ElseIf Me.Protocol = sckUDPProtocol Then
      ErrorCode = WSASendTo(m_hSocket, ByVal VarPtr(Bfr), 1&, cbDataSent, _
        Flags, ByVal VarPtr(m_RemoteAddr), LenB(m_RemoteAddr), ByVal VarPtr(OL), 0&)
    Else
      ' We should never be here.
      ErrorCode = WSAEPROTONOSUPPORT
      Debug.Assert 0
    End If
    
    If ErrorCode = SOCKET_ERROR Then
      ErrorCode = Err.LastDllError
    End If
    
    Select Case ErrorCode
    
    Case NO_ERROR
      ' Data written without waiting
      
      ' Keep track of total bytes outbound
      cbTotalSent = cbTotalSent + cbDataSent
    
    Case ERROR_IO_PENDING
      
      ' Need to wait on our event
      If WaitOnEvent(OL.hEvent, Me.SendTimeout, lpAbort:=VarPtr(m_Aborted)) > 0 Then
        ' Data is on its way

        ' Keep track of total bytes outbound
        cbTotalSent = cbTotalSent + cbDataSent

      Else
        ' oops; something went wrong
        Debug.Assert 0
        WsaRaiseError Err.LastDllError, Module, "Could not send data."

      End If

    Case Else
      ' oops; something went wrong
      Debug.Assert 0
      WsaRaiseError Err.LastDllError, "Could not send data."

    End Select

    ' Notify caller of write progress for each chunk until complete.
    ' This allows the caller to give feedback to the user (and to
    ' pull the carpet out from underneath us).
    RaiseEventEx iSendProgress, cbTotalSent, Bfr.len - cbTotalSent

  Loop While (cbTotalSent < Bfr.len)

  ' If synchronous then we notify caller that
  ' send completed via an event.
  If Synchronous Then
    ' Ignore callers errors in event
    On Error Resume Next
    RaiseEventEx iSendComplete
  End If

ExitLabel:
  On Error Resume Next
  
  If Not InvalidHandleValue(OL.hEvent) Then
    Call WSACloseEvent(OL.hEvent)
    OL.hEvent = INVALID_HANDLE_VALUE
  End If
  
  If ErrNum <> 0 Then
    ' If synchronous then we just raise errors back.
    ' Otherwise we notify the caller via the Error event.
    If Synchronous Then
      On Error GoTo 0
      ErrRaise ErrNum, ErrMsg
    Else
      RaiseEventEx iError, ErrNum, ErrMsg
    End If
  End If
  
  Exit Sub
  Resume
ErrHandler:
  ' Save error info
  ErrNum = Err.Number: ErrMsg = Err.Description
  Debug.Assert 0
  Resume ExitLabel

End Sub


Private Function GetSocket() As Long
'===========================================================================
' GetSocket - Creates a new socket, if one is not already created. It also
' creates the private window needed to hook socket-based events.
'===========================================================================

  Dim ErrNum      As Long
  Dim Failure     As Boolean
  
  If InvalidHandleValue(m_hSocket) Then
  
    Call CreatePrivateWindow
    
    ' Connect to the desired port
    If Me.Protocol = sckUDPProtocol Then
      m_hSocket = MWinsock.Socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)
    ElseIf Me.Protocol = sckTCPProtocol Then
      m_hSocket = MWinsock.Socket(AF_INET, SOCK_STREAM, IPPROTO_IP)
    Else
      Debug.Assert 0
      Err.Raise 5, Module, "Unexpected error. Unsupported protocol: " & Me.Protocol
    End If

    If m_hSocket = INVALID_SOCKET Then
      ErrNum = Err.LastDllError
      RemoveProp m_hWnd, App.EXEName
      DestroyWindow m_hWnd
      WsaRaiseError ErrNum, Module, "Could not create socket."
    End If
    
    State = sckOpen
      
  Else
    GetSocket = m_hSocket
  End If
  
  GetSocket = m_hSocket
    
End Function


Private Sub DestroySocket()
'===========================================================================
' DestroySocket - Destroys the socket an associated private window.
'===========================================================================
  
  Dim baBogus() As Byte
  
  If Not InvalidHandleValue(m_hSocket) Then
  
    On Error Resume Next
    
    ' We want to flush out the buffers if there
    ' is data that has not been read yet.
    If Me.BytesReceived > 0 Then
      Me.GetData baBogus
    End If
    
    ' Attempt to perform a valid shutdown
    DoSocketEvents
    If shutdown(m_hSocket, SD_BOTH) = SOCKET_ERROR Then
      If Err.LastDllError = WSAEINPROGRESS Then
        ' Best we can do is wait a few.
        ' Although we should never be here since
        ' we do not use a blocking call.
        Pause 3000
        DoSocketEvents
      End If
    End If

    Call CloseSocket(m_hSocket)
    m_hSocket = INVALID_HANDLE_VALUE
    State = sckClosed
    
  End If

End Sub

Public Function Initialize()
Attribute Initialize.VB_Description = "Hidden method reserved for MTS\r\n"
Attribute Initialize.VB_MemberFlags = "40"
  ' Hidden method reserved for MTS
  Class_Initialize
End Function

Public Function Terminate()
Attribute Terminate.VB_Description = "Hidden method reserved for MTS"
Attribute Terminate.VB_MemberFlags = "40"
  ' Hidden method reserved for MTS
  Class_Terminate
End Function

Private Sub Class_Initialize()
'===========================================================================
' Intialize - Initializes the socket library and this class for use.
'===========================================================================

  Dim ErrorCode       As Long

  ' Default protocol
  Me.Protocol = sckTCPProtocol
  
  ' Fetch info about socket version and initialize library.
  ' Throw error on failure.
  ErrorCode = WSAStartup(MakeWord(2, 2), m_WSAData)

  If ErrorCode <> NO_ERROR Then
    WsaRaiseError ErrorCode, Module
  ElseIf LoBite(m_WSAData.wVersion) < 2 Then
    ' Probably only version 1.1 installed.
    Call WSACleanup
    ErrRaise sckUnsupportedVersion
  End If
    
End Sub

Private Sub Class_Terminate()

  OutputDebugStringA "Terminating object instance."

  ' Be certain to decrement the library usage counter for this process.
  WSACleanup

  Call Destroy

End Sub


Private Sub RaiseEventEx(ByVal EventID As EventIDs, ParamArray Params() As Variant)
'===========================================================================
'   RaiseEventEx - Notifies the caller early bound if there is an event handler
'   in place, otherwise the caller is notified using the standard RaiseEvent.
'
'   NOTE: Be sure to qualify objects (with default properties) when
'   passing to variant parameters.
'===========================================================================

  ' Ignore host's errors
  On Error Resume Next
  
  If EventHandler Is Nothing Then
      
    OutputDebugStringA "RaiseEvent(" & EventID & ")"
    ' Standard events via RaiseEvent
    Select Case EventID
    Case iCloseConnection
      RaiseEvent CloseConnection
    Case iConnect
      RaiseEvent Connect
    Case iConnectionRequest
      RaiseEvent ConnectionRequest(Params(0))
    Case iDataArrival
      RaiseEvent DataArrival(Params(0))
    Case iError
      RaiseEvent Error(Params(0), Params(1))
    Case iSendComplete
      RaiseEvent SendComplete
    Case iSendProgress
      RaiseEvent SendProgress(Params(0), Params(1))
    Case Else
      Debug.Assert 0
    End Select
  
  Else
      
    ' Early bound events
    With EventHandler
        
      OutputDebugStringA "EventHandler(" & EventID & ")"
      Select Case EventID
      Case iCloseConnection
        .CloseConnection Me
      Case iConnect
        .Connect Me
      Case iConnectionRequest
        .ConnectionRequest Me, Params(0)
      Case iDataArrival
        .DataArrival Me, Params(0)
      Case iError
        .Error Me, Params(0), Params(1)
      Case iSendComplete
        .SendComplete Me
      Case iSendProgress
        .SendProgress Me, Params(0), Params(1)
      Case Else
        Debug.Assert 0
      End Select
    
    End With
      
  End If
  
ExitLabel:

  Exit Sub
  Resume
ErrHandler:
  Debug.Assert 0
  Resume ExitLabel
    
End Sub


Private Sub ErrRaise(ByVal Number As WinsockErrors, Optional ByVal Description As String)
'===========================================================================
' ErrRaise - Central routine to raise errors.
'===========================================================================

  On Error GoTo 0
  
  Select Case Number
  ' TODO: This should be mapped from WSA_OPERATION_ABORTED
  Case sckOpCanceled '1004            ' The operation was canceled.
    Err.Raise Number, Module, "The operation was canceled."
      
  ' -- Winsock control related errors
  Case sckBadState '40006
    Err.Raise Number, Module, "Wrong protocol or connection state for the requested transaction or request."
  
  Case sckInvalidArg ' 40014
    ' They should have used #5 Invalid Procedure Call
    ' but we will use sckInvalidArg to be compatible.
    Err.Raise Number, Module, "The argument passed to a function was not in the correct format or in the specified range."
  
  'Case sckSuccess = 40017
  '  Err.Raise Number, Module, "Successful."
  
  Case sckUnsupported ' 40018
    ' They should have used #13 Type Mismatch
    ' but we will use sckUnsupported to be compatible.
    Err.Raise Number, Module, "Unsupported variant type."
  
  Case sckInvalidOp ' 40020            '
    Err.Raise Number, Module, "Invalid operation at current state."
  
  Case sckOutOfRange ' 40021           '
    Err.Raise Number, Module, "Argument is out of range."
  
  Case sckWrongProtocol ' 40026        '
    Err.Raise Number, Module, "Wrong protocol for the requested transaction or request."
  
  ' -- Errors specific to this class
  Case sckUnsupportedVersion ' 40050   '
    Err.Raise Number, Module, "The installed version of Winsock is unsupported. Install Sockets version 2.0 or above."
  
  Case Else
      
    ' Use VBA defined error description if not supplied.
    If Len(Description) = 0 Then
      Description = VBA.Error$(Number)
    End If
    
    ' Throw the error
    Err.Raise Number, Module, Description
      
  End Select

End Sub

Public Function WindowProc(ByVal hWnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
Attribute WindowProc.VB_MemberFlags = "40"
'===========================================================================
' WindowProc - The WindowProc function is an application-defined function that processes
' messages sent to a window. The WNDPROC type defines a pointer to this callback function.
' WindowProc is a placeholder for the application-defined function name.
'
' hWnd            Handle to the window.
' uMsg            Specifies the message.
' wParam          Specifies additional message information. The contents of this parameter
'                 depend on the value of the uMsg parameter.
' lParam          Specifies additional message information. The contents of this parameter
'                 depend on the value of the uMsg parameter.
' RETURNS         The return value is the result of the message processing and depends on
'                 the message sent.
'===========================================================================

  Dim FD_Event    As FD_Events
  Dim ErrorCode   As SocketErrors
  Dim RequestID   As Long
  Dim ErrNum      As Long
  Dim ErrMsg      As String
  
  ' Trap ALL errors
  On Error GoTo ErrHandler
  
  Select Case uMsg
  Case WM_CLOSE
    
    ' better tear down the class if not already done
    RemoveProp m_hWnd, App.EXEName
    DestroyWindow m_hWnd
    m_hWnd = 0
      
  Case SocketMessage()
  
    FD_Event = LoWord(lParam)
    ErrorCode = HiWord(lParam)
    RequestID = wParam
    
    OutputDebugStringA "ErrorCode=" & ErrorCode & "; RequestID=" & RequestID
    
    Select Case FD_Event
    Case FD_ACCEPT
    
      OutputDebugStringA "FD_ACCEPT"
      
      ' There's an incoming connection
      If ErrorCode = 0 Then
  
        'If Not InvalidHandleValue(m_hSocket) Then
          ' Notify the host and allow them to call the
          ' Accept method. Ignore errors.
          RaiseEventEx iConnectionRequest, m_hSocket 'RequestID '
        'Else
        '    Debug.Assert 0
        'End If
      Else
          
        Call DestroySocket
        
        ' Notify caller of the write error.
        RaiseEventEx iError, _
          ErrorCode, _
          "Unable to accept incoming connection. " & LastDllErrorMsg(ErrorCode)
      
      End If
    
    ' Identify the FD_WRITE event
    Case FD_WRITE

      OutputDebugStringA "FD_WRITE"

      ' Socket is ready to receive data
      If ErrorCode = 0 Then

        ' Remove the FD_WRITE bit and "reselect" events to wait on.
        ' We do this because a caller could potentially call the
        ' SendData method during this write operation which would
        ' reselect with the FD_WRITE event -- this way we don't
        ' overwrite the FD_WRITE request.
        m_NetEvents = m_NetEvents And Not FD_WRITE

        If WSAAsyncSelect(m_hSocket, m_hWnd, SocketMessage(), m_NetEvents) = SOCKET_ERROR Then
          WsaRaiseError Err.LastDllError, Module
        End If

        ' Send the outgoing data asynchronously. Loop until
        ' the entire outbound async write queue is processed.
        Do While m_OutgoingData.Count > 0
          ' Send the data
          DataSend m_OutgoingData(1), HeapSize(GetProcessHeap(), 0, m_OutgoingData(1)), Synchronous:=True
          ' Release the memory allocated for this
          ' single data element item in the outbound queue.
          HeapFree GetProcessHeap(), HEAP_NO_SERIALIZE, m_OutgoingData(1)
          ' Pop data item from the queue
          If m_OutgoingData.Count > 0 Then m_OutgoingData.Remove 1
        Loop

      Else
          
        ' Notify caller of the write error.
        RaiseEventEx iError, _
          ErrorCode, _
          LastDllErrorMsg(ErrorCode)
      
      End If

        ' Identify the FD_READ event
    Case FD_READ
        
      OutputDebugStringA "FD_READ"
      
      ' We have received data
      If ErrorCode = 0 Then
          
        Dim nBytes      As Long
        nBytes = Me.BytesReceived
        If nBytes > 0 Then
          RaiseEventEx iDataArrival, nBytes
        End If
          
      Else
          
        ' Notify caller of the read error.
        RaiseEventEx iError, _
          ErrorCode, _
          LastDllErrorMsg(ErrorCode)
          
      End If
  
      ' Remove the FD_CONNECT bit and "reselect" events to wait on.
      ' We do this because we only connect once in this subroutine.
      m_NetEvents = m_NetEvents And Not FD_CONNECT
      If WSAAsyncSelect(m_hSocket, m_hWnd, SocketMessage(), m_NetEvents) = SOCKET_ERROR Then
        WsaRaiseError Err.LastDllError, Module
      End If
    
    ' Identify the FD_CONNECT event
    Case FD_CONNECT

      OutputDebugStringA "FD_CONNECT"
      
      ' we have received a connection
      If ErrorCode = 0 Then

        ' we have received a connection

        ' Notify the host and allow them to call the
        ' Accept method. Ignore errors.
        State = sckConnected
        
        RaiseEventEx iConnect

      Else
          
        Call DestroySocket
        
        ' Notify caller of the connect error.
        RaiseEventEx iError, _
          ErrorCode, _
          LastDllErrorMsg(ErrorCode)
          
      End If
    
        ' Identify the FD_CLOSE event
    Case FD_CLOSE
            
      OutputDebugStringA "FD_CLOSE(" & m_hSocket & ")"
      
      ' we have received a connection
      If ErrorCode <> 0 Then
          
        ' Notify caller of the error trying to close.
        RaiseEventEx iError, _
          ErrorCode, _
          LastDllErrorMsg(ErrorCode)

      End If

      Call DestroySocket
      RaiseEventEx iCloseConnection

    End Select
      
  Case Else
    ' who cares
  End Select

ExitLabel:
  WindowProc = 1

  Exit Function
  Resume
ErrHandler:
  ' We may error out if trying to select events on
  ' a socket that was closed by the caller during
  ' the processing of an event raised in this routine.
  If Err.Number = (sckNotSocket Or vbObjectError) Then Resume ExitLabel
  ' This can happen because the connection gets closed
  ' by the caller while processing a message.
  If InvalidHandleValue(m_hSocket) Then Resume ExitLabel
  Debug.Assert 0
  
  ' Save error info
  ErrNum = Err.Number
  ErrMsg = Err.Description
  
  ' Turn off error handler
  On Error Resume Next
  ' Notify caller of the error via an event
  RaiseEventEx iError, ErrNum, ErrMsg
  ' Restore error handler
  On Error GoTo ErrHandler
  ' Skip past the error
  Resume ExitLabel

End Function

Private Sub CreatePrivateWindow()
'===========================================================================
' CreatePrivateWindow - creates the window to receive private messages from
' the Windows' system tray.
'===========================================================================
    
  Dim WndCls      As WNDCLASS      ' Window class for this private window

  If CBool(IsWindow(m_hWnd)) Then Exit Sub

  ' Clear system error
  Err.Clear

  With WndCls
    ' We only set what we need:
    .hInstance = App.hInstance
    .lpszClassName = "WinsockClass"
    .lpfnwndproc = ProcAddress(AddressOf MWinsock.WindowProc)
  End With

  ' Register the window class before creating the window
  RegisterClass WndCls
  
  ' Create the window
  m_hWnd = CreateWindowEx(0, "WinsockClass", _
    "WinsockClass", WS_POPUP, 0, 0, 0, 0, 0, 0, App.hInstance, 0)

  ' Validate window
  If IsWindow(m_hWnd) = 0 Then
    ApiRaise Err.LastDllError, Module, "Could not create private window."
  End If
  
  ' Save weak reference to this object instance.
  ' The property name doesn't really matter since
  ' we own the window.
  Call SetProp(m_hWnd, App.EXEName, ObjPtr(Me))

End Sub

Private Function DoSocketEvents() As Boolean
'===========================================================================
' DoSocketEvents - Message pump for our private window.
'===========================================================================

  Const WM_QUIT = &H12
  Const PM_REMOVE = &H1
  
  Dim theMsg As MSG
  
  If CBool(PeekMessage(theMsg, m_hWnd, 0&, 0&, PM_REMOVE)) Then
    
    If theMsg.Message = WM_QUIT Then
      Exit Function
    End If
    
    TranslateMessage theMsg
    DispatchMessage theMsg
  
  End If
  
  DoSocketEvents = True
  
End Function
