Attribute VB_Name = "MCommonUI"
Option Explicit


Public Sub SetControlState( _
  ByVal Control As Control, _
  ByVal IsEnabled As Boolean, _
  Optional ByVal LockInputOnly As Boolean = False)
'===========================================================================
' SetControlState - Changes the input state and appearance of a control to
' "Enabled" or "Disabled".
'===========================================================================
    
  ' Sanity check
  If Control Is Nothing Then Exit Sub
    
  Select Case TypeName(Control)
  Case "Label"

    ' For a Label, just change the forecolor
    SetControlColor Control, _
      IsEnabled Or LockInputOnly
    
  Case "TextBox", "ComboBox"
    
    ' Change Locked state & background color to match
    Control.Locked = Not IsEnabled
    SetControlColor Control, IsEnabled
    
  Case "ListBox"

    ' Listbox doesn't have a "Locked" property.
    Control.Enabled = IsEnabled Or LockInputOnly
    SetControlColor Control, IsEnabled
  
  Case "Frame", "PictureBox", "CheckBox", _
       "OptionButton", "CommandButton"
  
    ' Change state of known non-input controls.
    ' Ignore error if "Enabled" not exported.
    On Error Resume Next
    Control.Enabled = IsEnabled Or LockInputOnly
    
  Case Else
  
    ' Change state of other non-input controls
    ' Ignore error if "Enabled" not exported.
    On Error Resume Next
    Control.Enabled = IsEnabled Or LockInputOnly
  
  End Select

End Sub

Public Sub SetControlColor(Control As Control, IsEnabled As Boolean)
'===========================================================================
' SetControlColor - Changes the appearance of a control to appear
' "Enabled" or "Disabled".
'===========================================================================

  Dim BackColor   As OLE_COLOR
  Dim ForeColor   As OLE_COLOR

  ' Sanity check
  If Control Is Nothing Then Exit Sub
  
  Select Case TypeName(Control)
  Case "Label", "Frame", "PictureBox", "CheckBox", _
       "OptionButton", "CommandButton"

    If IsEnabled Then
      Control.ForeColor = vbWindowText
    Else
      Control.ForeColor = vbGrayText 'vbInactiveCaptionText
    End If
    
  Case "TextBox", "ComboBox", "ListBox"
    
    If IsEnabled Then
      Control.BackColor = vbWindowBackground
    Else
      Control.BackColor = vbButtonFace
    End If

  Case Else

    If IsEnabled Then
      Control.BackColor = vbWindowBackground
    Else
      Control.BackColor = vbButtonFace
    End If
  
  End Select

End Sub

Public Sub SetControlChildState( _
  ByVal Control As Control, _
  IsEnabled As Boolean, _
  Optional UnNest As Boolean = True, _
  Optional LockInputOnly As Boolean = False)
'===========================================================================
' SetControlChildState - Changes the state of a group of controls contained
' within supplied container.
'===========================================================================
  
  Dim Container   As Object
  Dim Ctl         As Control
  Dim hwnd        As Long

  ' Sanity check
  If Control Is Nothing Then Exit Sub
  
  ' Changes the Enabled or Locked property according to the current state.
  ' Also changes the Background color.
  
  ' Handle errors inline
  On Error Resume Next
  
  ' loop thru each control
  For Each Ctl In Control.Parent.Controls
    
    Set Container = Ctl.Container             ' save container reference
    ' NOTE: An error may have raised if it was a menu control,
    ' or some other control that doesn't export "Container" property.
    
    If Not Container Is Nothing Then          ' control has a container
      
      ' Get window handle for container
      hwnd = Ctl.Container.hwnd
      
      If hwnd = Control.hwnd Then
        ' it's a match
        
        Select Case TypeName(Ctl)
        
        Case "Frame", "PictureBox"
          ' These are known control containers

          If UnNest Then
            SetControlChildState Ctl, IsEnabled, UnNest, LockInputOnly
          End If
        
        Case Else

          ' If it's a child, set the state
          SetControlState Ctl, IsEnabled, LockInputOnly

        End Select

      End If
    
    End If
  
    ' Cleanup
    Set Container = Nothing
    hwnd = 0
  
  Next Ctl
  
End Sub


