Attribute VB_Name = "MWinsock"
'===========================================================================
'
' MWinsock Module
'
' Support module for the Winsock.Cls class. This is not a port of the entire
' winsock headers, but rather, only that which is needed by the class.
'
'===========================================================================
'
' Author:             Monte Hansen [monte@killervb.com]
' Dependencies:       None
'
'===========================================================================
'
' * Copyright  1999-2002 by Monte Hansen, All Rights Reserved Worldwide *
'
' Monte Hansen (The  Author)  grants a royalty-free right to use,  modify,
' and distribute this code (The Code) in compiled form,  provided that you
' agree that The Author has no warranty,  obligations or liability for The
' Code.  You may distribute  The Code  among peers but may not sell it, or
' distribute  it  on  any  electronic  or  physical media such  as  floppy
' diskettes,  compact disks,  bulletin boards,  web  sites,  and the like,
' without first obtaining The Author's consent.
'
' When distributing  The Code  among peers,  it is  respectfully requested
' that it be distributed  as is,  but at  no time  shall it be distributed
' without the copyright notice hereinabove.
'
' This license is subject to change without notice. The current license can
' be found at this address: http://www.killervb.com/kvbLicense.aspx
'
' A credit would be nice if you use The Code ;>
'
'===========================================================================
Option Explicit

'===========================================================================
'   Constants
'===========================================================================
Private Const Module          As String = "MWinsock"

Private m_Initialized         As Boolean        ' True if WSAStartup called
Private m_WSAData             As WSAData        ' Socket version information
Private m_WM_WINSOCK_MSG      As Long           ' Custom message used for socket msg handler

Public Enum SockProtocols
  IPPROTO_IP = 0                          ' /* dummy for IP */
  IPPROTO_ICMP = 1                        ' /* control message protocol */
  IPPROTO_IGMP = 2                        ' /* internet group management protocol */
  IPPROTO_GGP = 3                         ' /* gateway^2 (deprecated) */
  IPPROTO_TCP = 6                         ' /* tcp */
  IPPROTO_PUP = 12                        ' /* pup */
  IPPROTO_UDP = 17                        ' /* user datagram protocol */
  IPPROTO_IDP = 22                        ' /* xns idp */
  IPPROTO_ND = 77                         ' /* UNOFFICIAL net disk proto */
  IPPROTO_RAW = 255                       ' /* raw IP packet */
  IPPROTO_MAX = 256
End Enum

Public Enum FD_Events
  FD_READ = 2 ^ 0
  FD_WRITE = 2 ^ 1
  FD_OOB = 2 ^ 2
  FD_ACCEPT = 2 ^ 3
  FD_CONNECT = 2 ^ 4
  FD_CLOSE = 2 ^ 5
  FD_QOS = 2 ^ 6
  FD_GROUP_QOS = 2 ^ 7
  FD_ROUTING_INTERFACE_CHANGE = 2 ^ 8
  FD_ADDRESS_LIST_CHANGE = 2 ^ 9
  FD_ALL_EVENTS = (2 ^ 10) - 1
End Enum

Public Enum FD_ErrorIndexes
  FD_READ_BIT = 0
  FD_WRITE_BIT = 1
  FD_OOB_BIT = 2
  FD_ACCEPT_BIT = 3
  FD_CONNECT_BIT = 4
  FD_CLOSE_BIT = 5
  FD_QOS_BIT = 6
  FD_GROUP_QOS_BIT = 7
  FD_ROUTING_INTERFACE_CHANGE_BIT = 8
  FD_ADDRESS_LIST_CHANGE_BIT = 9
End Enum

'/*
' * Level number for (get/set)sockopt() to apply to socket itself.
' */
Public Const SOL_SOCKET = &HFFFF          '/* options for socket level */


Public Enum MessageFlags
  MSG_OOB = &H1&                          '/* process out-of-band data */
  MSG_PEEK = &H2&                         '/* peek at incoming message */
  MSG_DONTROUTE = &H4&                    '/* send without using routing tables */
  MSG_PARTIAL = &H8000&                   '/* partial send or recv for message xport */
'/*
'* Winsock 2 extension -- new flags for WSASend(), WSASendTo(), WSARecv() and
'*                          WSARecvFrom()
'*/
  MSG_INTERRUPT = &H10&                   '/* send/recv in the interrupt context */
  MSG_MAXIOVLEN = 16&
End Enum

'/*
' * Winsock 2 extension -- manifest constants for shutdown()
' */
Public Const SD_RECEIVE = &H0&
Public Const SD_SEND = &H1&
Public Const SD_BOTH = &H2&

Public Const INVALID_SOCKET = -1&
Public Const SOCKET_ERROR = -1&
Public Const NO_ERROR = 0&

Public Const WSADESCRIPTION_LEN = 256&
Public Const WSASYS_STATUS_LEN = 128&
Public Const WSAPROTOCOL_LEN = 255&
Public Const MAX_PROTOCOL_CHAIN = 7&

Public Const FIONREAD = &H4004667F          ' /* get # bytes to read */
Public Const FIONBIO = &H8004667E           ' /* set/clear non-blocking i/o */
Public Const FIOASYNC = &H8004667D          ' /* set/clear async i/o */
Public Const DONT_BLOCK_CALL = 1&           ' Custom arg for use with FIONBIO
Public Const BLOCK_CALL = 1&                ' Custom arg for use with FIONBIO
Public Const FD_MAX_EVENTS = 10&

'/*
' * Definitions of bits in internet address integers.
' * On subnets, the decomposition of addresses to host and net parts
' * is done according to subnet mask, not the masks here.
' */
Public Const INADDR_ANY = 0&
Public Const INADDR_LOOPBACK = &H7F000001
Public Const INADDR_BROADCAST = &HFFFFFFFF
Public Const INADDR_NONE = &HFFFFFFFF
Public Const ADDR_ANY = INADDR_ANY

' /*
'  * All Windows Sockets error constants are biased by WSABASEERR from
'  * the "normal"
'  */
Public Const WSABASEERR = 10000&

Public Enum SocketErrors

  ' /*
  '  * Windows Sockets definitions of regular Microsoft C error constants
  '  */
  WSAEINTR = (WSABASEERR + 4)
  WSAEBADF = (WSABASEERR + 9)
  WSAEACCES = (WSABASEERR + 13)
  WSAEFAULT = (WSABASEERR + 14)
  WSAEINVAL = (WSABASEERR + 22)
  WSAEMFILE = (WSABASEERR + 24)
  
  ' /*
  '  * Windows Sockets definitions of regular Berkeley error constants
  '  */
  WSAEWOULDBLOCK = (WSABASEERR + 35)
  WSAEINPROGRESS = (WSABASEERR + 36)
  WSAEALREADY = (WSABASEERR + 37)
  WSAENOTSOCK = (WSABASEERR + 38)
  WSAEDESTADDRREQ = (WSABASEERR + 39)
  WSAEMSGSIZE = (WSABASEERR + 40)
  WSAEPROTOTYPE = (WSABASEERR + 41)
  WSAENOPROTOOPT = (WSABASEERR + 42)
  WSAEPROTONOSUPPORT = (WSABASEERR + 43)
  WSAESOCKTNOSUPPORT = (WSABASEERR + 44)
  WSAEOPNOTSUPP = (WSABASEERR + 45)
  WSAEPFNOSUPPORT = (WSABASEERR + 46)
  WSAEAFNOSUPPORT = (WSABASEERR + 47)
  WSAEADDRINUSE = (WSABASEERR + 48)
  WSAEADDRNOTAVAIL = (WSABASEERR + 49)
  WSAENETDOWN = (WSABASEERR + 50)
  WSAENETUNREACH = (WSABASEERR + 51)
  WSAENETRESET = (WSABASEERR + 52)
  WSAECONNABORTED = (WSABASEERR + 53)
  WSAECONNRESET = (WSABASEERR + 54)
  WSAENOBUFS = (WSABASEERR + 55)
  WSAEISCONN = (WSABASEERR + 56)
  WSAENOTCONN = (WSABASEERR + 57)
  WSAESHUTDOWN = (WSABASEERR + 58)
  WSAETOOMANYREFS = (WSABASEERR + 59)
  WSAETIMEDOUT = (WSABASEERR + 60)
  WSAECONNREFUSED = (WSABASEERR + 61)
  WSAELOOP = (WSABASEERR + 62)
  WSAENAMETOOLONG = (WSABASEERR + 63)
  WSAEHOSTDOWN = (WSABASEERR + 64)
  WSAEHOSTUNREACH = (WSABASEERR + 65)
  WSAENOTEMPTY = (WSABASEERR + 66)
  WSAEPROCLIM = (WSABASEERR + 67)
  WSAEUSERS = (WSABASEERR + 68)
  WSAEDQUOT = (WSABASEERR + 69)
  WSAESTALE = (WSABASEERR + 70)
  WSAEREMOTE = (WSABASEERR + 71)
  
  ' /*
  '  * Extended Windows Sockets error constant definitions
  '  */
  WSASYSNOTREADY = (WSABASEERR + 91)
  WSAVERNOTSUPPORTED = (WSABASEERR + 92)
  WSANOTINITIALISED = (WSABASEERR + 93)
  WSAEDISCON = (WSABASEERR + 101)
  WSAENOMORE = (WSABASEERR + 102)
  WSAECANCELLED = (WSABASEERR + 103)
  WSAEINVALIDPROCTABLE = (WSABASEERR + 104)
  WSAEINVALIDPROVIDER = (WSABASEERR + 105)
  WSAEPROVIDERFAILEDINIT = (WSABASEERR + 106)
  WSASYSCALLFAILURE = (WSABASEERR + 107)
  WSASERVICE_NOT_FOUND = (WSABASEERR + 108)
  WSATYPE_NOT_FOUND = (WSABASEERR + 109)
  WSA_E_NO_MORE = (WSABASEERR + 110)
  WSA_E_CANCELLED = (WSABASEERR + 111)
  WSAEREFUSED = (WSABASEERR + 112)
  
  ' /*
  '  * Error return codes from gethostbyname() and gethostbyaddr()
  '  * (when using the resolver). Note that these errors are
  '  * retrieved via WSAGetLastError() and must therefore follow
  '  * the rules for avoiding clashes with error numbers from
  '  * specific implementations or language run-time systems.
  '  * For this reason the codes are based at WSABASEERR+1001.
  '  * Note also that [WSA]NO_ADDRESS is defined only for
  '  * compatibility purposes.
  '  */
  
  ' /* Authoritative Answer: Host not found'  */
  WSAHOST_NOT_FOUND = (WSABASEERR + 1001)
  
  ' /* Non-Authoritative: Host not found, or SERVERFAIL'  */
  WSATRY_AGAIN = (WSABASEERR + 1002)
  
  ' /* Non-recoverable errors, FORMERR, REFUSED, NOTIMP'  */
  WSANO_RECOVERY = (WSABASEERR + 1003)
  
  ' /* Valid name, no data record of requested type'  */
  WSANO_DATA = (WSABASEERR + 1004)
  
  ' /*
  '  * Define QOS related error return codes
  '  *
  '  */
  WSA_QOS_RECEIVERS = (WSABASEERR + 1005)
           ' /* at least one Reserve has arrived'  */
  WSA_QOS_SENDERS = (WSABASEERR + 1006)
           ' /* at least one Path has arrived'  */
  WSA_QOS_NO_SENDERS = (WSABASEERR + 1007)
           ' /* there are no senders'  */
  WSA_QOS_NO_RECEIVERS = (WSABASEERR + 1008)
           ' /* there are no receivers'  */
  WSA_QOS_REQUEST_CONFIRMED = (WSABASEERR + 1009)
           ' /* Reserve has been confirmed'  */
  WSA_QOS_ADMISSION_FAILURE = (WSABASEERR + 1010)
           ' /* error due to lack of resources'  */
  WSA_QOS_POLICY_FAILURE = (WSABASEERR + 1011)
           ' /* rejected for administrative reasons - bad credentials'  */
  WSA_QOS_BAD_STYLE = (WSABASEERR + 1012)
           ' /* unknown or conflicting style'  */
  WSA_QOS_BAD_OBJECT = (WSABASEERR + 1013)
           ' /* problem with some part of the filterspec or providerspecific
           '  * buffer in general'  */
  WSA_QOS_TRAFFIC_CTRL_ERROR = (WSABASEERR + 1014)
           ' /* problem with some part of the flowspec'  */
  WSA_QOS_GENERIC_ERROR = (WSABASEERR + 1015)
           ' /* general error'  */
  WSA_QOS_ESERVICETYPE = (WSABASEERR + 1016)
           ' /* invalid service type in flowspec'  */
  WSA_QOS_EFLOWSPEC = (WSABASEERR + 1017)
           ' /* invalid flowspec'  */
  WSA_QOS_EPROVSPECBUF = (WSABASEERR + 1018)
           ' /* invalid provider specific buffer'  */
  WSA_QOS_EFILTERSTYLE = (WSABASEERR + 1019)
           ' /* invalid filter style'  */
  WSA_QOS_EFILTERTYPE = (WSABASEERR + 1020)
           ' /* invalid filter type'  */
  WSA_QOS_EFILTERCOUNT = (WSABASEERR + 1021)
           ' /* incorrect number of filters'  */
  WSA_QOS_EOBJLENGTH = (WSABASEERR + 1022)
           ' /* invalid object length'  */
  WSA_QOS_EFLOWCOUNT = (WSABASEERR + 1023)
           ' /* incorrect number of flows'  */
  WSA_QOS_EUNKOWNPSOBJ = (WSABASEERR + 1024)
           ' /* unknown object in provider specific buffer'  */
  WSA_QOS_EPOLICYOBJ = (WSABASEERR + 1025)
           ' /* invalid policy object in provider specific buffer'  */
  WSA_QOS_EFLOWDESC = (WSABASEERR + 1026)
           ' /* invalid flow descriptor in the list'  */
  WSA_QOS_EPSFLOWSPEC = (WSABASEERR + 1027)
           ' /* inconsistent flow spec in provider specific buffer'  */
  WSA_QOS_EPSFILTERSPEC = (WSABASEERR + 1028)
           ' /* invalid filter spec in provider specific buffer'  */
  WSA_QOS_ESDMODEOBJ = (WSABASEERR + 1029)
           ' /* invalid shape discard mode object in provider specific buffer'  */
  WSA_QOS_ESHAPERATEOBJ = (WSABASEERR + 1030)
           ' /* invalid shaping rate object in provider specific buffer'  */
  WSA_QOS_RESERVED_PETYPE = (WSABASEERR + 1031)
           ' /* reserved policy element in provider specific buffer'  */

End Enum ' SocketErrors

'/*
' * Types
' */
Public Const SOCK_STREAM = 1&               ' /* stream socket */
Public Const SOCK_DGRAM = 2&                ' /* datagram socket */
Public Const SOCK_RAW = 3&                  ' /* raw-protocol interface */
Public Const SOCK_RDM = 4&                  ' /* reliably-delivered message */
Public Const SOCK_SEQPACKET = 5&            ' /* sequenced packet stream */

'/*
' * Option flags per-socket.
' */
Public Const SO_DEBUG = &H1                 ' /* turn on debugging info recording */
Public Const SO_ACCEPTCONN = &H2            ' /* socket has had listen() */
Public Const SO_REUSEADDR = &H4             ' /* allow local address reuse */
Public Const SO_KEEPALIVE = &H8             ' /* keep connections alive */
Public Const SO_DONTROUTE = &H10            ' /* just use interface addresses */
Public Const SO_BROADCAST = &H20            ' /* permit sending of broadcast msgs */
Public Const SO_USELOOPBACK = &H40          ' /* bypass hardware when possible */
Public Const SO_LINGER = &H80               ' /* linger on close if data present */
Public Const SO_OOBINLINE = &H100           ' /* leave received OOB data in line */

' NOTE: I'm not 100% sure about this declare
' since the header performs a cast to an
' unsigned integer:
'public const SO_DONTLINGER   (u_int)(~SO_LINGER)
Public Const SO_DONTLINGER = Not SO_LINGER

'/*
' * Additional options.
' */
Public Const SO_SNDBUF = &H1001             ' /* send buffer size */
Public Const SO_RCVBUF = &H1002             ' /* receive buffer size */
Public Const SO_SNDLOWAT = &H1003           ' /* send low-water mark */
Public Const SO_RCVLOWAT = &H1004           ' /* receive low-water mark */
Public Const SO_SNDTIMEO = &H1005           ' /* send timeout */
Public Const SO_RCVTIMEO = &H1006           ' /* receive timeout */
Public Const SO_ERROR = &H1007              ' /* get error status and clear */
Public Const SO_TYPE = &H1008               ' /* get socket type */


'/*
' * Maximum queue length specifiable by listen.
' */
Public Const SOMAXCONN = &H7FFFFFFF

'/*
' * TCP options.
' */
Public Const TCP_NODELAY = &H1&

'/*
' * Address families.
' */
Public Const AF_UNSPEC = 0&                 ' /* unspecified */

'/*
' * Although  AF_UNSPEC  is  defined for backwards compatibility, using
' * AF_UNSPEC for the "af" parameter when creating a socket is STRONGLY
' * DISCOURAGED.    The  interpretation  of  the  "protocol"  parameter
' * depends  on the actual address family chosen.  As environments grow
' * to  include  more  and  more  address families that use overlapping
' * protocol  values  there  is  more  and  more  chance of choosing an
' * undesired address family when AF_UNSPEC is used.
' */
Public Const AF_UNIX = 1                  ' /* local to host (pipes, portals) */
Public Const AF_INET = 2                  ' /* internetwork: UDP, TCP, etc. */
Public Const AF_IMPLINK = 3               ' /* arpanet imp addresses */
Public Const AF_PUP = 4                   ' /* pup protocols: e.g. BSP */
Public Const AF_CHAOS = 5                 ' /* mit CHAOS protocols */
Public Const AF_NS = 6                    ' /* XEROX NS protocols */
Public Const AF_IPX = AF_NS               ' /* IPX protocols: IPX, SPX, etc. */
Public Const AF_ISO = 7                   ' /* ISO protocols */
Public Const AF_OSI = AF_ISO              ' /* OSI is ISO */
Public Const AF_ECMA = 8                  ' /* european computer manufacturers */
Public Const AF_DATAKIT = 9               ' /* datakit protocols */
Public Const AF_CCITT = 10                ' /* CCITT protocols, X.25 etc */
Public Const AF_SNA = 11                  ' /* IBM SNA */
Public Const AF_DECnet = 12               ' /* DECnet */
Public Const AF_DLI = 13                  ' /* Direct data link interface */
Public Const AF_LAT = 14                  ' /* LAT */
Public Const AF_HYLINK = 15               ' /* NSC Hyperchannel */
Public Const AF_APPLETALK = 16            ' /* AppleTalk */
Public Const AF_NETBIOS = 17              ' /* NetBios-style addresses */
Public Const AF_VOICEVIEW = 18            ' /* VoiceView */
Public Const AF_FIREFOX = 19              ' /* Protocols from Firefox */
Public Const AF_UNKNOWN1 = 20             ' /* Somebody is using this! */
Public Const AF_BAN = 21                  ' /* Banyan */
Public Const AF_ATM = 22                  ' /* Native ATM Services */
Public Const AF_INET6 = 23                ' /* Internetwork Version 6 */
Public Const AF_CLUSTER = 24              ' /* Microsoft Wolfpack */
Public Const AF_12844 = 25                ' /* IEEE 1284.4 WG AF */
Public Const AF_IRDA = 26                 ' /* IrDA */
Public Const AF_NETDES = 28               ' /* Network Designers OSI & gateway enabled protocols */
Public Const AF_MAX = 29

'===============================================================================
' Types
'===============================================================================

Public Type GUID
  Data1 As Long
  Data2 As Integer
  Data3 As Integer
  Data4(7) As Byte
End Type

Public Type OVERLAPPED
  Internal As Long
  InternalHigh As Long
  Offset As Long
  OffsetHigh As Long
  hEvent As Long
End Type

'/*
' * Select uses arrays of SOCKETs.  These macros manipulate such
' * arrays.  FD_SETSIZE may be defined by the user before including
' * this file, but the default here should be >= 64.
' *
' * CAVEAT IMPLEMENTOR and USER: THESE MACROS AND TYPES MUST BE
' * INCLUDED IN WINSOCK2.H EXACTLY AS SHOWN HERE.
' */
Public Const FD_SETSIZE = 64
Public Type fd_set
  fd_count            As Long             ' /* how many are SET? */
  fd_array(1 To FD_SETSIZE) As Long       ' /* an array of SOCKETs */
End Type

'/*
' * Winsock 2 extension -- WSABUF and QOS struct, include qos.h
' * to pull in FLOWSPEC and related definitions
' */

Type WSABUF
  len                 As Long             ' /* the length of the buffer */
  Buf                 As Long             ' /* the pointer to the buffer */
End Type

'/*
' * Structure used in select() call, taken from the BSD file sys/time.h.
' */
Public Type timeval
  tv_sec              As Long             ' /* seconds */
  tv_usec             As Long             ' /* and microseconds */
End Type

'/*
' * Internet address (old style... should be updated)
' */
Public Type in_addr
  s_addr              As Long             ' see SDK for layout
End Type

'/*
' * Socket address, internet style.
' */
Public Type sockaddr_in
  sin_family          As Integer          ' Address family (must be AF_INET).
  sin_port            As Integer          ' IP port.
  sin_addr            As in_addr          ' IP address.
  sin_zero(7)         As Byte             ' Padding to make structure the same size as SOCKADDR.
End Type

'/*
' * Structure used by kernel to store most addresses.
' */
Public Type sockaddr
  sa_family           As Integer          ' /* address family */
  sa_data(13)         As Byte             ' /* up to 14 bytes of direct address */
End Type

'/*
' * Structures returned by network data base library, taken from the
' * BSD file netdb.h.  All addresses are supplied in host order, and
' * returned in network order (suitable for use in system calls).
' */
Type hostent
  h_name              As Long             ' /* official name of host */
  h_aliases           As Long             ' /* alias list */
  h_addrtype          As Integer          ' /* host address type */
  h_length            As Integer          ' /* length of address */
  h_addr_list         As Long             ' /* list of addresses */
End Type

Public Type WSAData
  wVersion             As Integer
  wHighVersion         As Integer
  szDescription        As String * 257    'WSADESCRIPTION_LEN+1
  szSystemStatus       As String * 129    'WSASYS_STATUS_LEN+1
  iMaxSockets          As Integer
  iMaxUdpDg            As Integer
  lpVendorInfo         As Long
End Type

'/*
' * Winsock 2 extension -- data type for WSAEnumNetworkEvents()
' */
Type WSANETWORKEVENTS
  lNetworkEvents      As Long
  iErrorCode(0 To FD_MAX_EVENTS - 1) As Long
End Type

Type WSAPROTOCOLCHAIN
  ChainLen            As Long             ' /* the length of the chain,     */
                                          ' /* length = 0 means layered protocol, */
                                          ' /* length = 1 means base protocol, */
                                          ' /* length > 1 means protocol chain */
  ChainEntries(1 To MAX_PROTOCOL_CHAIN) As Long ' /* a list of dwCatalogEntryIds */
End Type

Type WSAPROTOCOL_INFO
  dwServiceFlags1     As Long
  dwServiceFlags2     As Long
  dwServiceFlags3     As Long
  dwServiceFlags4     As Long
  dwProviderFlags     As Long
  ProviderId          As GUID
  dwCatalogEntryId    As Long
  ProtocolChain       As WSAPROTOCOLCHAIN
  iVersion            As Long
  iAddressFamily      As Long
  iMaxSockAddr        As Long
  iMinSockAddr        As Long
  iSocketType         As Long
  iProtocol           As Long
  iProtocolMaxOffset  As Long
  iNetworkByteOrder   As Long
  iSecurityScheme     As Long
  dwMessageSize       As Long
  dwProviderReserved  As Long
  szProtocol(1 To WSAPROTOCOL_LEN + 1) As Byte
End Type

'===============================================================================
' Declares
'===============================================================================

Public Declare Function Accept Lib "wsock32.dll" _
  Alias "accept" (ByVal s As Long, nameaddr As Any, addrlen As Long) As Long
  
Public Declare Function Bind Lib "wsock32.dll" _
  Alias "bind" (ByVal s As Long, nameaddr As Any, ByVal namelen As Long) As Long
  
Public Declare Function CloseSocket Lib "wsock32.dll" _
  Alias "closesocket" (ByVal s As Long) As Long

Public Declare Function Connect Lib "wsock32.dll" _
  Alias "connect" (ByVal s As Long, nameaddr As Any, ByVal namelen As Long) As Long
  
Public Declare Function gethostbyaddr Lib "wsock32.dll" _
  (addr As Long, ByVal addrlen As Long, ByVal addrtype As Long) As Long
  
Public Declare Function gethostbyname Lib "wsock32.dll" _
  (ByVal HostName As String) As Long
  
Public Declare Function gethostname Lib "wsock32.dll" _
  (ByVal HostName As String, ByVal namelen As Long) As Long
  
Public Declare Function getpeername Lib "wsock32.dll" _
  (ByVal s As Long, nameaddr As Any, namelen As Long) As Long
  
Public Declare Function getprotobyname Lib "wsock32.dll" _
  (ByVal ProtocolName As String) As Long
  
Public Declare Function getprotobynumber Lib "wsock32.dll" _
  (ByVal Number As Long) As Long
  
Public Declare Function getservbyname Lib "wsock32.dll" _
  (ByVal servname As String, ByVal proto As String) As Long
  
Public Declare Function getservbyport Lib "wsock32.dll" _
  (ByVal port As Long, ByVal proto As String) As Long
  
Public Declare Function getsockname Lib "wsock32.dll" _
  (ByVal s As Long, nameaddr As Any, namelen As Long) As Long
  
Public Declare Function getsockopt Lib "wsock32.dll" _
  (ByVal s As Long, ByVal level As Long, ByVal optname As Long, optval As Any, optlen As Long) As Long
  
Public Declare Function htonl Lib "wsock32.dll" _
  (ByVal hostlong As Long) As Long
  
Public Declare Function htons Lib "wsock32.dll" _
  (ByVal hostshort As Long) As Integer

Public Declare Function inet_addr Lib "wsock32.dll" _
  (ByVal cp As String) As Long

Public Declare Function inet_ntoa Lib "wsock32.dll" _
  (ByVal inaddr As Long) As Long

Public Declare Function ioctlsocket Lib "wsock32.dll" _
  (ByVal s As Long, ByVal cmd As Long, argp As Long) As Long

Public Declare Function Listen Lib "wsock32.dll" _
  Alias "listen" (ByVal s As Long, ByVal backlog As Long) As Long

Public Declare Function ntohl Lib "wsock32.dll" _
  (ByVal netlong As Long) As Long

Public Declare Function ntohs Lib "wsock32.dll" _
  (ByVal netshort As Long) As Integer

Public Declare Function recv Lib "wsock32.dll" _
  (ByVal s As Long, Buf As Any, ByVal buflen As Long, ByVal Flags As Long) As Long

Public Declare Function recvfrom Lib "wsock32.dll" _
  (ByVal s As Long, Buf As Any, ByVal buflen As Long, ByVal Flags As Long, addrfrom As Any, fromlen As Long) As Long

Public Declare Function select_ Lib "wsock32.dll" Alias "select" _
  (ByVal nfds As Long, readfds As fd_set, writefds As fd_set, exceptfds As fd_set, TimeOut As timeval) As Long

Public Declare Function send Lib "wsock32.dll" _
  (ByVal s As Long, Buf As Any, ByVal buflen As Long, ByVal Flags As Long) As Long

Public Declare Function sendto_ Lib "wsock32.dll" _
  (ByVal s As Long, Buf As Any, ByVal buflen As Long, ByVal Flags As Long, addrto As Any, ByVal tolen As Long) As Long

Public Declare Function setsockopt Lib "wsock32.dll" _
  (ByVal s As Long, ByVal level As Long, ByVal optname As Long, optval As Any, ByVal optlen As Long) As Long

Public Declare Function shutdown Lib "wsock32.dll" _
  (ByVal s As Long, ByVal how As Long) As Long

Public Declare Function Socket Lib "wsock32.dll" _
  Alias "socket" (ByVal af As Long, ByVal sockettype As Long, ByVal Protocol As Long) As Long

Public Declare Function WSASocket Lib "wsock32.dll" _
  (ByVal af As Long, _
   ByVal addr_type As Long, _
   ByVal Protocol As Long, _
   lpProtocolInfo As WSAPROTOCOL_INFO, _
   ByVal reserved As Long, _
   ByVal dwFlags As Long) As Long

Public Declare Function WSAStartup Lib "wsock32.dll" _
  (ByVal wVersionRequested As Integer, lpWSAData As WSAData) As Long

Public Declare Function WSACleanup Lib "wsock32.dll" () As Long

Public Declare Function WSACloseEvent Lib "ws2_32.dll" (ByVal hEvent As Long) As Long

Public Declare Function WSACreateEvent Lib "ws2_32.dll" () As Long

Public Declare Function WSADuplicateSocket Lib "ws2_32.dll" _
  (ByVal s As Long, ByVal dwProcessId As Long, lpProtocolInfo As WSAPROTOCOL_INFO) As Long

Public Declare Function WSAEventSelect Lib "ws2_32.dll" _
  (ByVal s As Long, ByVal hEventObject As Long, ByVal lNetworkEvents As FD_Events) As Long

Public Declare Function WSAResetEvent Lib "ws2_32.dll" _
  (ByVal hEventObject As Long) As Long

Public Declare Function WSASend Lib "ws2_32.dll" _
  (ByVal s As Long, lpBuffers As Any, ByVal dwBufferCount As Long, _
   lpNumberOfBytesSent As Long, ByVal dwFlags As Long, _
   lpOverlapped As Any, ByVal lpCompletionRoutine As Long) As Long

Public Declare Function WSASendTo Lib "ws2_32.dll" _
  (ByVal s As Long, lpBuffers As Any, ByVal dwBufferCount As Long, _
   lpNumberOfBytesSent As Long, ByVal dwFlags As Long, _
   lpToAddr As Any, ByVal iToAddrLen As Long, _
   lpOverlapped As Any, ByVal lpCompletionRoutine As Long) As Long

Public Declare Function WSARecv Lib "ws2_32.dll" _
  (ByVal s As Long, lpBuffers As Any, ByVal dwBufferCount As Long, _
   lpNumberOfBytesRecvd As Long, dwFlags As Long, lpOverlapped As Any, _
   ByVal lpCompletionRoutine As Long) As Long
  
Public Declare Function WSAEnumNetworkEvents Lib "ws2_32.dll" _
  (ByVal s As Long, ByVal hEventObject As Long, lNetworkEvents As WSANETWORKEVENTS) As Long

Public Declare Sub WSASetLastError Lib "wsock32.dll" (ByVal iError As Long)

Public Declare Function WSAGetLastError Lib "wsock32.dll" () As Long

Public Declare Function WSAIsBlocking Lib "wsock32.dll" () As Long

Public Declare Function WSAUnhookBlockingHook Lib "wsock32.dll" () As Long

Public Declare Function WSASetBlockingHook Lib "wsock32.dll" (ByVal lpBlockFunc As Long) As Long

Public Declare Function WSACancelBlockingCall Lib "wsock32.dll" () As Long

Public Declare Function WSAAsyncGetServByName Lib "wsock32.dll" _
  (ByVal hWnd As Long, ByVal wMsg As Long, ByVal serv_name As String, _
   ByVal proto As String, Buf As Any, ByVal buflen As Long) As Long

Public Declare Function WSAAsyncGetServByPort Lib "wsock32.dll" _
  (ByVal hWnd As Long, ByVal wMsg As Long, ByVal port As Long, _
   ByVal proto As String, Buf As Any, ByVal buflen As Long) As Long
   
Public Declare Function WSAAsyncGetProtoByName Lib "wsock32.dll" _
  (ByVal hWnd As Long, ByVal wMsg As Long, ByVal proto_name As String, _
   Buf As Any, ByVal buflen As Long) As Long

Public Declare Function WSAAsyncGetProtoByNumber Lib "wsock32.dll" _
  (ByVal hWnd As Long, ByVal wMsg As Long, ByVal Number As Long, _
   Buf As Any, ByVal buflen As Long) As Long
   
Public Declare Function WSAAsyncGetHostByName Lib "wsock32.dll" _
  (ByVal hWnd As Long, ByVal wMsg As Long, ByVal host_name As String, _
   Buf As Any, ByVal buflen As Long) As Long

Public Declare Function WSAAsyncGetHostByAddr Lib "wsock32.dll" _
  (ByVal hWnd As Long, ByVal wMsg As Long, addr As Long, _
   ByVal addr_len As Long, ByVal addr_type As Long, Buf As Any, _
   ByVal buflen As Long) As Long

Public Declare Function WSACancelAsyncRequest Lib "wsock32.dll" _
  (ByVal hAsyncTaskHandle As Long) As Long

Public Declare Function WSAAsyncSelect Lib "wsock32.dll" _
  (ByVal s As Long, ByVal hWnd As Long, ByVal wMsg As Long, ByVal lEvent As Long) As Long

Public Declare Function WSARecvEx Lib "wsock32.dll" _
  (ByVal s As Long, Buf As Any, ByVal buflen As Long, ByVal Flags As Long) As Long

Public Declare Function WSAGetOverlappedResult Lib "ws2_32.dll" _
  (ByVal s As Long, _
   lpOverlapped As Any, _
   lpcbTransfer As Long, _
   ByVal fWait As Long, _
   lpdwFlags As Long) As Long

Declare Function WSAConnect Lib "ws2_32.dll" _
  (ByVal s As Long, _
   name As sockaddr_in, _
   ByVal namelen As Long, _
   lpCallerData As Long, _
   lpCalleeData As Long, _
   lpSQOS As Long, _
   lpGQOS As Long) As Long

Public Declare Function KillTimer Lib "user32" _
  (ByVal hWnd As Long, ByVal nIDEvent As Long) As Long
  
Public Declare Function SetTimer Lib "user32" _
  (ByVal hWnd As Long, ByVal nIDEvent As Long, _
   ByVal uElapse As Long, ByVal lpTimerFunc As Long) As Long

Public Declare Function CreateEvent Lib "kernel32" Alias "CreateEventA" _
  (lpEventAttributes As Any, ByVal bManualReset As Long, _
   ByVal bInitialState As Long, ByVal lpName As String) As Long
  
Public Declare Function GetOverlappedResult Lib "kernel32" _
  (ByVal hFile As Long, lpOverlapped As Any, _
   lpNumberOfBytesTransferred As Long, ByVal bWait As Long) As Long

Private Declare Sub CopyMemoryAny Lib "kernel32" Alias "RtlMoveMemory" (lpDest As Any, lpSource As Any, ByVal nBytes As Long)
Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (ByVal lpDest As Long, ByVal lpSource As Long, ByVal nBytes As Long)
Private Declare Function GetProp Lib "user32" Alias "GetPropA" (ByVal hWnd As Long, ByVal lpString As String) As Long
Private Declare Function SetProp Lib "user32" Alias "SetPropA" (ByVal hWnd As Long, ByVal lpString As String, ByVal hData As Long) As Long
Private Declare Sub OutputDebugStringA Lib "kernel32" (ByVal lpOutputString As String)
Private Declare Function RegisterWindowMessage Lib "user32" Alias "RegisterWindowMessageA" (ByVal lpString As String) As Long

Public Sub WsaRaiseError(Optional ErrorCode As Long, Optional ByVal Source As String, Optional DescPrefix As String)

  If ErrorCode = 0 Then ErrorCode = Err.LastDllError
  ApiRaise ErrorCode, Module, DescPrefix
    
End Sub

Public Function AddressToName(ByVal Address As Long) As String
'===============================================================================
' AddressToName - Returns the name of the host machine given it's address.
'===============================================================================

  Dim lpHostEnt       As Long
  Dim Host            As hostent
  
  ' Sanity check
  If Address = 0 Then Exit Function
  
  ' Get name of host given its IP address.
  lpHostEnt = gethostbyaddr(Address, 4, AF_INET)
  If lpHostEnt = 0 Then
    WsaRaiseError Err.LastDllError, Module
  End If
  
  ' Copy to our local HOSTENT structure
  CopyMemory ByVal VarPtr(Host), ByVal lpHostEnt, Len(Host)
  
  ' Convert pointer to string & return the name of the host
  AddressToName = PtrToStrA(Host.h_name)

End Function


Public Function AddressToIP(ByVal Address As Long) As String
'===============================================================================
' AddressToIP - Returns the host machine's dotted IP address given it's
' address.
'===============================================================================

  Dim lpIP            As Long
  Dim Bytes(3)        As Byte
  Dim i               As Long
  
  ' Sanity check
  If Address = 0 Then Exit Function
  
  ' Convert address to a dotted string
  lpIP = inet_ntoa(Address)

  ' Convert pointer to string & return the IP string
  AddressToIP = PtrToStrA(lpIP)

End Function


Public Function NameToAddress(ByVal name As String) As Long
'===============================================================================
' NameToAddress - Returns the address of the host machine given it's name.
'===============================================================================

  Dim lpHostEnt       As Long
  Dim Host            As hostent
  
  ' Get name of host given its IP address.
  lpHostEnt = gethostbyname(name)
  If lpHostEnt = 0 Then
    WsaRaiseError Err.LastDllError, Module
  End If
  
  ' Copy to our local HOSTENT structure
  CopyMemory ByVal VarPtr(Host), ByVal lpHostEnt, Len(Host)
  
  ' Dereference our pointer to a pointer to an (IP) address and return it
  If Host.h_length = 4 Then
    CopyMemory ByVal VarPtr(NameToAddress), ByVal Host.h_addr_list, 4
    CopyMemory ByVal VarPtr(NameToAddress), ByVal NameToAddress, 4
  Else
    Debug.Assert 0
    ' we are not prepared for anything else.
  End If

End Function

Public Function AnyToAddress(ByVal HostOrIpString As String) As Long
'===============================================================================
' AnyToAddress - Returns the address of the host machine given it's name or IP.
'
' This function is needed due to a strange thing I have seen under NT. A netbios
' hostname can appear to be a valid address reprsentation that can result in NT
' sending it outbound never to be seen again.
'===============================================================================
  
  If IsDottedAddress(HostOrIpString) Then
    
    ' Convert IP address string to an internet address structure.
    AnyToAddress = inet_addr(HostOrIpString)

  Else
    
    ' RemoteHost not supplied in dotted IP form.
    ' Therefore we attempt to resolve.
      
    ' Lookup the host name and convert it
    ' to an internet address structure
    AnyToAddress = NameToAddress(HostOrIpString)
  
  End If
  
End Function

Public Function StartWinsock(Optional sDescription As String) As Boolean
    
  If Not m_Initialized Then

    m_Initialized = WSAStartup(MakeWord(2, 2), m_WSAData)
    sDescription = TrimNulls(m_WSAData.szDescription)

  End If
  
  StartWinsock = m_Initialized

End Function

Sub EndWinsock()
    
  Dim Result      As Long
  
  ' If calls are blocking then kill em
  If WSAIsBlocking() Then
    Result = WSACancelBlockingCall()
  End If
  
  ' Decrement ref counter
  Result = WSACleanup()
  m_Initialized = False

End Sub


Public Function WindowProc(ByVal hWnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
'===============================================================================
' WindowProc - The WindowProc function is an application-defined function that
' processes messages sent to a window. The WNDPROC type defines a pointer to
' this callback function.
'
' hWnd            Handle to the window.
' uMsg            Specifies the message.
' wParam          Specifies additional message information. The contents of
'                 this parameter depend on the value of the uMsg parameter.
' lParam          Specifies additional message information. The contents of
'                 this parameter depend on the value of the uMsg parameter.
' RETURNS         The return value is the result of the message processing
'                 and depends on the message sent.
'===============================================================================

  On Error GoTo ErrHandler

  Dim Socket      As Winsock
  Dim lpWinsock   As Long
  
  ' Don't bother processing messages other than
  ' our registered socket message.
  If uMsg = SocketMessage() Then
      
    lpWinsock = GetProp(hWnd, App.EXEName)
    Set Socket = PtrToObj(lpWinsock)
    If Socket Is Nothing Then GoTo ExitLabel
    WindowProc = Socket.WindowProc(hWnd, uMsg, wParam, lParam)
      
  Else
  
    'WindowProc = DefWindowProc(hWnd, uMsg, wParam, lParam)
  
  End If
  
ExitLabel:
  WindowProc = 1
  Exit Function
  Resume
ErrHandler:
  OutputDebugStringA "WindowProc Error: " & Err.Description
  Debug.Assert 0
  ' we should never be here

End Function


Public Function PtrToObj(ByVal lpObject As Long) As Object

  Dim Obj         As Winsock

  On Error GoTo ErrHandler

  ' Get the object pointer attached to this window's user data
  If InvalidHandleValue(lpObject) Then Exit Function

  ' Create the weak reference
  Call CopyMemoryAny(Obj, lpObject, 4)

  ' Return the object reference
  If Not Obj Is Nothing Then
    Set PtrToObj = Obj
  End If

  ' Destroy the weak reference
  Call CopyMemoryAny(Obj, 0&, 4)

  Exit Function
  Resume
ErrHandler:
  OutputDebugStringA "PtrToObj Failed. " & Err.Description
  On Error GoTo ErrHandler
  Set PtrToObj = Nothing
  Resume Next
    
End Function


Public Property Get SocketMessage() As Long
'===============================================================================
' SocketMessage - This is the private window message used to handle Winsock
' specific events.
'===============================================================================

  If m_WM_WINSOCK_MSG = 0 Then
    m_WM_WINSOCK_MSG = RegisterWindowMessage(App.EXEName & ".SocketMessage")
  End If

  SocketMessage = m_WM_WINSOCK_MSG
    
End Property


Public Function IsDottedAddress(ByVal Address As String) As Boolean
'===============================================================================
' IsDottedAddress - This functions identifies if a DNS address is in hostname
' format or dotted ip.
'===============================================================================

  Dim vSegments As Variant
  Dim i         As Long
  
  ' Split address into segments
  vSegments = VBA.Split(Address, ".")
  
  ' Exit if not 4 segments
  If Not IsArray(vSegments) Then Exit Function
  If Not UBound(vSegments) = 3 Then Exit Function
  
  ' Ensure each segment is a valid byte value
  For i = LBound(vSegments) To UBound(vSegments)
    
    If Not IsNumeric(vSegments(i)) Then Exit Function
    
    Select Case Val(vSegments(i))
    Case 0 To 255
      ' move on to next segment if in valid range.
    Case Else
      Exit Function
    End Select
    
  Next i
  
  ' Return a hero if we are here.
  IsDottedAddress = True
  
End Function

